#include "G4INCLAbla07Interface.hh"
#include "G4INCLLogger.hh"
#include "G4INCLParticleTable.hh"
#include "G4INCLGlobals.hh"
#include "G4INCLRandom.hh"
#include <cstring>
#include <cassert>
#include <algorithm>

namespace ABLA07CXX {

  const double Abla07Interface::fmp = 938.27231;
  const double Abla07Interface::fmn = 939.56563;

  Abla07Interface::Abla07Interface(G4INCL::Config *conf) :
    IDeExcitation(conf),
    theConfig(conf),
    verboseEvent(conf->getVerboseEvent()) {

      const int dataFilePathLength = 1000;
      char dataFilePath[dataFilePathLength+1];
      std::strncpy(dataFilePath, theConfig->getABLA07DataFilePath().c_str(), dataFilePathLength);
      const int len = std::strlen(dataFilePath);
      // Terminate the string with a space so that the Fortran subroutine knows
      // how to locate the end of the path
      // (Yes, filenames with spaces are not supported)
      dataFilePath[len] = ' ';
      dataFilePath[len+1] = 0;

      const int Zt = theConfig->getTargetZ();
      int At;
      if(theConfig->isNaturalTarget()) {
        At = G4INCL::ParticleTable::getNaturalIsotopicDistribution(Zt).getIsotopes().back().theA;
      } else
        At = theConfig->getTargetA();

      const G4INCL::ParticleType t = theConfig->getProjectileType();
      const int Ap = G4INCL::ParticleTable::getMassNumber(t);
      const int Zp = G4INCL::ParticleTable::getChargeNumber(t);
      double Ep = theConfig->getProjectileKineticEnergy();

      // Initialize ABLA07
      abramain_.at=At;
      abramain_.zt=Zt;
      abramain_.eap=Ep;
      abramain_.ap=At;
      abramain_.zp=Zt;
      proj_.eap1=Ep;
      proj_.ap1=At;
      proj_.zp1=Zt;
      abrainit_(dataFilePath,dataFilePathLength);

      // ABLA07 thinks in inverse kinematics. Swap target and projectile
      abramain_.ap=At;
      abramain_.zp=Zt;
      abramain_.at=Ap;
      abramain_.zt=Zp;
    }

  Abla07Interface::~Abla07Interface() {
  }

  void Abla07Interface::deExciteRemnant(G4INCL::EventInfo *eventInfo, const int i) {
    int particleIndex = eventInfo->nParticles;
    int eventNumber = eventInfo->eventNumber;

    double EStarRem = eventInfo->EStarRem[i];
    int ARem = eventInfo->ARem[i];
    int ZRem = eventInfo->ZRem[i];
    double pxRem = eventInfo->pxRem[i];
    double pyRem = eventInfo->pyRem[i];
    double pzRem = eventInfo->pzRem[i];
    double JRem = eventInfo->JRem[i];

    // Call the de-excitation subroutine
    rn07_(&eventNumber, &ARem, &ZRem, &EStarRem, &pxRem, &pyRem, &pzRem, &JRem);
    // Fill the common block with results
    ablantuple_(&eventNumber,&verboseEvent);

    for(int j = 0; j < zablaout_.nniv; ++j) { // Copy ABLA result to the EventInfo
      int Apart = zablaout_.nna[j];
      int Zpart = zablaout_.nnz[j];
      eventInfo->A[particleIndex] = Apart;
      eventInfo->Z[particleIndex] = Zpart;
      eventInfo->emissionTime[particleIndex] = -1.0;

      const double px = zablaout_.ppx[j];
      const double py = zablaout_.ppy[j];
      const double pz = zablaout_.ppz[j];
      const double plab = std::sqrt(px*px+py*py+pz*pz);
      eventInfo->px[particleIndex] = px;
      eventInfo->py[particleIndex] = py;
      eventInfo->pz[particleIndex] = pz;

      double pznorm = pz/plab;
      eventInfo->theta[particleIndex] = G4INCL::Math::toDegrees(G4INCL::Math::arcCos(pznorm));
      eventInfo->phi[particleIndex] = G4INCL::Math::toDegrees(std::atan2(py,px));

      eventInfo->EKin[particleIndex] = zablaout_.eekin[j];;
      eventInfo->origin[particleIndex] = i; // Origin: De-excitation
      eventInfo->history.push_back(historyIntToString(zablaout_.ww[j])); // history
      particleIndex++;
    }
    eventInfo->nParticles = particleIndex;
  }

  std::string Abla07Interface::historyIntToString(const int hist) {
    std::stringstream ss;
    ss << hist;
    std::string histString(ss.str());
    std::transform(histString.begin(),
        histString.end(),
        histString.begin(),
        ABLA07CXX::Abla07Interface::HistoryStepIntToChar());
    return histString;
  }

  char Abla07Interface::HistoryStepIntToChar::historyDictionary[Abla07Interface::HistoryStepIntToChar::historyDictionarySize];

  Abla07Interface::HistoryStepIntToChar::HistoryStepIntToChar() {
    // Initialise the history dictionary
    /* From GEMINI++'s CHistory.h:
       enum HistoryStepType {
       Evaporation=1,
       EvaporationResidue,
       Multifragmentation,
       AsymmetricFissionLight,
       AsymmetricFissionHeavy,
       SymmetricFissionLight,
       SymmetricFissionHeavy,
       SaddleToScission,
       NonStatistical
       */
    historyDictionary[0] = 0;   // nil
    historyDictionary[1] = 'e'; // evaporation
    historyDictionary[2] = 'E'; // evaporation residue
    historyDictionary[3] = 'm'; // multifragmentation
    historyDictionary[4] = 'a'; // asymmetric fission, light partner
    historyDictionary[5] = 'A'; // asymmetric fission, heavy partner
    historyDictionary[6] = 'f'; // fission, light partner
    historyDictionary[7] = 'F'; // fission, heavy partner
    historyDictionary[8] = 's'; // saddle-to-scission emission
    historyDictionary[9] = 'n'; // non-statistical emission (decay)
  }

  char Abla07Interface::HistoryStepIntToChar::operator()(const char step){
    std::stringstream ss;
    ss << step;
    int istep;
    ss >> istep;
    assert(istep>=0 && istep<historyDictionarySize);
    return historyDictionary[istep];
  }

}

float abla07userrng_() {
  return ((float)G4INCL::Random::shoot());
}
