//
// ********************************************************************
// * License and Disclaimer                                           *
// *                                                                  *
// * The  Geant4 software  is  copyright of the Copyright Holders  of *
// * the Geant4 Collaboration.  It is provided  under  the terms  and *
// * conditions of the Geant4 Software License,  included in the file *
// * LICENSE and available at  http://cern.ch/geant4/license .  These *
// * include a list of copyright holders.                             *
// *                                                                  *
// * Neither the authors of this software system, nor their employing *
// * institutes,nor the agencies providing financial support for this *
// * work  make  any representation or  warranty, express or implied, *
// * regarding  this  software system or assume any liability for its *
// * use.  Please see the license in the file  LICENSE  and URL above *
// * for the full disclaimer and the limitation of liability.         *
// *                                                                  *
// * This  code  implementation is the result of  the  scientific and *
// * technical work of the GEANT4 collaboration.                      *
// * By using,  copying,  modifying or  distributing the software (or *
// * any work based  on the software)  you  agree  to acknowledge its *
// * use  in  resulting  scientific  publications,  and indicate your *
// * acceptance of all terms of the Geant4 Software license.          *
// ********************************************************************
//
// $Id: G4INCLVFermiBreakUp.cc,v 1.5 2006-06-29 20:13:13 gunter Exp $
//
// Hadronic Process: Nuclear De-excitations
// by V. Lara (Nov 1998)
//
// Modifications:
// 01.04.2011 General cleanup by V.Ivanchenko - more clean usage of static
// 23.04.2011 V.Ivanchenko: make this class to be responsible for
//            selection of decay channel and decay

#include <set>

#include "G4INCLFermiConfigurationList.hh"
#include "G4INCLFermiFragmentsPool.hh"
#include "G4INCLKinematicsUtils.hh"

const double G4INCLFermiConfigurationList::Kappa = 6.0;
const double G4INCLFermiConfigurationList::r0 = 1.3;

G4INCLFermiConfigurationList::G4INCLFermiConfigurationList()
{
  thePool = G4INCLFermiFragmentsPool::Instance(); 
  Coef = 0.6*(G4INCL::PhysicalConstants::eSquared/r0)/G4INCL::Math::pow13(1+int(Kappa));
  ConstCoeff = std::pow(r0/G4INCL::PhysicalConstants::hc,3)*Kappa*std::sqrt(2.0/G4INCL::Math::pi)/3.0;

  // 16 is the max number
  nmax = 50;
  NormalizedWeights.resize(nmax,0.0);
}

G4INCLFermiConfigurationList::~G4INCLFermiConfigurationList()
{}

double 
G4INCLFermiConfigurationList::CoulombBarrier(
  const std::vector<const G4INCLVFermiFragment*>& conf)
{
  //  Calculates Coulomb Barrier (MeV) for given channel with K fragments.
  int SumA = 0;
  int SumZ = 0;
  double CoulombEnergy = 0.;
  size_t nn = conf.size();
  for (size_t i=0; i<nn; ++i) {
    int z = conf[i]->GetZ();
    int a = conf[i]->GetA();
    CoulombEnergy += double(z*z)/G4INCL::Math::pow13(a);
    SumA += a;
    SumZ += z;
  }
  CoulombEnergy -= SumZ*SumZ/G4INCL::Math::pow13(SumA);
  return -Coef * CoulombEnergy;
}

double 
G4INCLFermiConfigurationList::DecayProbability(int A, double TotalE, 
					   G4INCLFermiConfiguration* conf)
  // Decay probability  for a given channel with K fragments
{
  // A: Atomic Weight
  // TotalE: Total energy of nucleus

  double KineticEnergy = TotalE; // MeV
  double ProdMass = 1.0;
  double SumMass = 0.0;
  double S_n = 1.0;
  std::set<int>      combSet;
  std::multiset<int> combmSet;

  const std::vector<const G4INCLVFermiFragment*> flist = 
    conf->GetFragmentList();

  // Number of fragments
  int K = flist.size();

  for (int i=0; i<K; ++i) {
    int a = flist[i]->GetA();
    combSet.insert(a);
    combmSet.insert(a);
    double mass = flist[i]->GetFragmentMass();
    ProdMass *= mass;
    SumMass += mass;
    // Spin factor S_n
    S_n *= flist[i]->GetPolarization();
    KineticEnergy -= mass + flist[i]->GetExcitationEnergy();
  }

  // Check that there is enough energy to produce K fragments
  KineticEnergy -= CoulombBarrier(flist);
  if (KineticEnergy <= 0.0) { return 0.0; }

  double MassFactor = ProdMass/SumMass;
  MassFactor *= std::sqrt(MassFactor);  
  
  // This is the constant (doesn't depend on nucleus) part
  double Coeff = std::pow(ConstCoeff*A, K-1);

  //JMQ 111009 Bug fixed: gamma function for odd K was wrong  by a factor 2
  // new implementation explicitely according to standard properties of Gamma function
  // Calculation of 1/Gamma(3(k-1)/2)
  double Gamma = 1.0;
  //  double arg = 3.0*(K-1)/2.0 - 1.0;
  //  while (arg > 1.1) 
  //    {
  //      Gamma *= arg; 
  //      arg--;
  //    }
  //  if ((K-1)%2 == 1) Gamma *= std::sqrt(pi);

  if ((K-1)%2 != 1)

    {
      double arg = 3.0*(K-1)/2.0 - 1.0;
      while (arg > 1.1) 
	{
	  Gamma *= arg; 
	  arg--;
	}
    }
  else   { 
    double 	n= 3.0*K/2.0-2.0;
    double arg2=2*n-1;
    while (arg2>1.1)
      {
	Gamma *= arg2;
	arg2  -= 2;
      }
    Gamma *= std::sqrt(G4INCL::Math::pi)/std::pow(2,n);
  }
  // end of new implementation of Gamma function
  
  
  // Permutation Factor G_n
  double G_n = 1.0;
  for (std::set<int>::iterator itr = combSet.begin(); 
       itr != combSet.end(); ++itr)
    {
      for (int ni = combmSet.count(*itr); ni > 1; ni--) { G_n *= ni; }
    }

  double Weight = Coeff * MassFactor * (S_n / G_n) / Gamma;
  Weight *=  std::sqrt(std::pow(KineticEnergy,3*(K-1)))/KineticEnergy;

  return Weight; 
}

G4INCL::ParticleVector*
G4INCLFermiConfigurationList::GetFragments(const G4INCL::Cluster& theNucleus)
{
  // Calculate Momenta of K fragments
  double M = G4INCL::KinematicsUtils::invariantMass(theNucleus.getEnergy(), theNucleus.getMomentum());
  const std::vector<const G4INCLVFermiFragment*>* conf = 
    SelectConfiguration(theNucleus.getZ(), 
			theNucleus.getA(), M);


  G4INCL::ParticleVector* theResult = new G4INCL::ParticleVector();
  size_t nn = conf->size();
  if(1 >= nn) {
    theResult->push_back(new G4INCL::Cluster(theNucleus));
    delete conf;
    return theResult; 
  }

  G4INCL::ThreeVector boostVector = theNucleus.boostVector();
  std::vector<double> mr;
  mr.reserve(nn);
  for(size_t i=0; i<nn; ++i) {
    mr.push_back( (*conf)[i]->GetTotalEnergy() );
  }
  std::vector<G4INCL::LorentzVector*>* mom = thePhaseSpace.Decay(M,mr);
  if(!mom) { 
    delete conf;
    return theResult; 
  }

  size_t nmom = mom->size();  

  // Go back to the Lab Frame
  if(0 < nmom) { 
    for (size_t j=0; j<nmom; ++j) {
      G4INCL::LorentzVector* FourMomentum = (*mom)[j];
    
      // Lorentz boost
      FourMomentum->boost(boostVector);
      
      G4INCL::ParticleVector* fragment = (*conf)[j]->GetFragment(*FourMomentum);
 
      size_t nfrag = fragment->size();
      for (size_t k=0; k<nfrag; ++k) { theResult->push_back((*fragment)[k]); }
      delete fragment;
      delete (*mom)[j];
    }
  }
  
  delete mom;
  delete conf;
  return theResult;
}

const std::vector<const G4INCLVFermiFragment*>* 
G4INCLFermiConfigurationList::SelectConfiguration(int Z, int A, double mass)
{
  std::vector<const G4INCLVFermiFragment*>* res = 
    new std::vector<const G4INCLVFermiFragment*>;  
  const std::vector<G4INCLFermiConfiguration*>* conflist = 
    thePool->GetConfigurationList(Z, A, mass);
  if(!conflist) { return res; }
  size_t nn = conflist->size();
  if(0 < nn) { 
    size_t idx = 0;
    if(1 < nn) {
      if(nn > nmax) {
        nmax = nn;
	NormalizedWeights.resize(nmax,0.0);
      }
      double prob = 0.0;
      for(size_t i=0; i<nn; ++i) { 
	prob += DecayProbability(A, mass, (*conflist)[i]);
	NormalizedWeights[i] = prob;
      }
      prob *= G4INCL::Random::shoot();
      for(idx=0; idx<nn; ++idx) { 
	if(NormalizedWeights[idx] >= prob) { break; }
      }
    }
    const std::vector<const G4INCLVFermiFragment*> flist = 
      (*conflist)[idx]->GetFragmentList();
    size_t nf = flist.size();
    for(size_t i=0; i<nf; ++i) { res->push_back(flist[i]); }
    //std::cout << "FermiBreakUp: " << nn << " conf; selected idx= " 
    //	 << idx << "  Nprod= " << nf << std::endl; 
  }
  delete conflist;
  return res;
}
