#include "G4INCLGEMINIXXInterface.hh"
#include "G4INCLParticleTable.hh"
#include "G4INCLGlobals.hh"
#include "G4INCLRandom.hh"
#include "CNucleus.h"
#include "CHistory.h"
#include <iostream>
#include <cstdlib>
#include <cassert>
#include <string>
#include <cstring>

double geminixxuserrng() {
  return G4INCL::Random::shoot();
}

const double G4INCLGEMINIXXInterface::theSpeedOfLight = 29.9792458; // cm/ns

G4INCLGEMINIXXInterface::G4INCLGEMINIXXInterface(G4INCL::Config *conf) :
  IDeExcitation(conf)
{
  std::string aString =
    "GINPUT=" +
    conf->getGEMINIXXDataFilePath();
  size_t len = aString.length()+1;
  theAssignment = new char[len];
  strncpy(theAssignment, aString.c_str(), len);
#if !defined(NDEBUG) && !defined(INCLXX_IN_GEANT4_MODE)
  int err =
#endif
    putenv(theAssignment);
#if !defined(NDEBUG) && !defined(INCLXX_IN_GEANT4_MODE)
  assert(err==0);
#endif
}

G4INCLGEMINIXXInterface::~G4INCLGEMINIXXInterface() {
  delete[] theAssignment;
}

void G4INCLGEMINIXXInterface::deExciteRemnant(G4INCL::EventInfo *eventInfo, const int i) {
  int particleIndex = eventInfo->nParticles;

  const int iZ = eventInfo->ZRem[i];
  const int iA = eventInfo->ARem[i];
  const float fEx = eventInfo->EStarRem[i];
  const float fJ = eventInfo->JRem[i];
  const float theRemnantMass = G4INCL::ParticleTable::getTableMass(iA,iZ) + fEx;
  const float theRemnantEnergy = eventInfo->EKinRem[i] + theRemnantMass;
  const float vx = theSpeedOfLight * eventInfo->pxRem[i] / theRemnantEnergy;
  const float vy = theSpeedOfLight * eventInfo->pyRem[i] / theRemnantEnergy;
  const float vz = theSpeedOfLight * eventInfo->pzRem[i] / theRemnantEnergy;
  const float jModule = std::sqrt(
                                  std::pow(eventInfo->jxRem[i],(float)2.)
                                  + std::pow(eventInfo->jyRem[i],(float)2.)
                                  + std::pow(eventInfo->jzRem[i],(float)2.)
                                 );
  float thetaJ, phiJ;
  if(jModule>0.) {
    thetaJ = G4INCL::Math::arcCos(eventInfo->jzRem[i]/jModule);
    phiJ = std::atan2(eventInfo->jyRem[i], eventInfo->jxRem[i]);
  } else {
    thetaJ = (eventInfo->jzRem[i]>=0.) ? 0. : G4INCL::Math::pi;
    phiJ = 0.;
  }

  CNucleus CN(iZ,iA,fEx,fJ);
  CN.setVelocityCartesian(vx,vy,vz);
  CAngle CNangle(thetaJ,phiJ);
  CN.setSpinAxis(CNangle);
  CN.decay();

  int nGEMINIXXProducts;
  //check to see if gemini was successfully able to decay the fragment
  if (CN.abortEvent) {
    nGEMINIXXProducts = 0;
  } else {

    CHistory theHistoryClass(&CN);

    nGEMINIXXProducts = CN.getNumberOfProducts();
    CNucleus* product;

    for(int j=0; j<nGEMINIXXProducts; j++)
    {
      product = CN.getProducts(j);
      eventInfo->A[particleIndex] = product->iA;
      eventInfo->Z[particleIndex] = product->iZ;
      eventInfo->emissionTime[particleIndex] = product->getTime();
      float *momentum = product->getMomentumVector();
      eventInfo->px[particleIndex] = momentum[0];
      eventInfo->py[particleIndex] = momentum[1];
      eventInfo->pz[particleIndex] = momentum[2];
      eventInfo->EKin[particleIndex] = product->getKE();
      const CAngle theAngle = product->getAngleDegrees();
      eventInfo->theta[particleIndex] = theAngle.theta;
      eventInfo->phi[particleIndex] = theAngle.phi;
      eventInfo->origin[particleIndex] = i; // Origin: De-excitation
      eventInfo->history.push_back(theHistoryClass.getHistoryAsString(product));
      particleIndex++;
    }
  }
  CN.reset();
  eventInfo->nParticles = particleIndex;
}

