/** \file G4INCLCoulombNonRelativistic.cc
 * \brief Class for non-relativistic Coulomb distortion.
 *
 * \date 14 February 2011
 * \author Davide Mancusi
 */

#include "G4INCLCoulombNonRelativistic.hh"
#include "G4INCLGlobals.hh"

namespace G4INCL {

  ParticleEntryAvatar *CoulombNonRelativistic::bringToSurface(Particle * const p, Nucleus * const n) const {
    // No distortion for neutral particles
    if(p->getZ()!=0) {
      const bool success = coulombDeviation(p, n);
      if(!success) // transparent
        return NULL;
    }

    // Rely on the CoulombNone slave to compute the straight-line intersection
    // and actually bring the particle to the surface of the nucleus
    return theCoulombNoneSlave.bringToSurface(p,n);
  }

  IAvatarList CoulombNonRelativistic::bringToSurface(Cluster * const c, Nucleus * const n) const {
    // Neutral clusters?!
    assert(c->getZ()>0);

    // Perform the actual Coulomb deviation
    const bool success = coulombDeviation(c, n);
    if(!success) {
      return IAvatarList();
    }

    // Rely on the CoulombNone slave to compute the straight-line intersection
    // and actually bring the particle to the surface of the nucleus
    return theCoulombNoneSlave.bringToSurface(c,n);
  }

  void CoulombNonRelativistic::distortOut(ParticleList const &pL,
      Nucleus const * const nucleus) const {

    for(ParticleIter particle=pL.begin(), e=pL.end(); particle!=e; ++particle) {

      const int Z = (*particle)->getZ();
      if(Z == 0) continue;

      const double tcos=1.-0.000001;

      const double et1 = PhysicalConstants::eSquared * nucleus->getZ();
      const double transmissionRadius =
        nucleus->getDensity()->getTransmissionRadius(*particle);

      const ThreeVector position = (*particle)->getPosition();
      ThreeVector momentum = (*particle)->getMomentum();
      const double r = position.mag();
      const double p = momentum.mag();
      const double cosTheta = position.dot(momentum)/(r*p);
      if(cosTheta < 0.999999) {
        const double sinTheta = std::sqrt(1.-cosTheta*cosTheta);
        const double eta = et1 * Z / (*particle)->getKineticEnergy();
        if(eta > transmissionRadius-0.0001) {
          // If below the Coulomb barrier, radial emission:
          momentum = position * (p/r);
          (*particle)->setMomentum(momentum);
        } else {
          const double b0 = 0.5 * (eta + std::sqrt(eta*eta +
                4. * std::pow(transmissionRadius*sinTheta,2)
                * (1.-eta/transmissionRadius)));
          const double bInf = std::sqrt(b0*(b0-eta));
          const double thr = std::atan(eta/(2.*bInf));
          double uTemp = (1.-b0/transmissionRadius) * std::sin(thr) +
            b0/transmissionRadius;
          if(uTemp>tcos) uTemp=tcos;
          const double thd = Math::arcCos(cosTheta)-Math::piOverTwo + thr +
            Math::arcCos(uTemp);
          const double c1 = std::sin(thd)*cosTheta/sinTheta + std::cos(thd);
          const double c2 = -p*std::sin(thd)/(r*sinTheta);
          const ThreeVector newMomentum = momentum*c1 + position*c2;
          (*particle)->setMomentum(newMomentum);
        }
      }
    }
  }

  double CoulombNonRelativistic::maxImpactParameter(ParticleSpecies const &p, const double kinE,
                                                    Nucleus const * const n) const {
    const double theMinimumDistance = minimumDistance(p, kinE, n);
    double rMax = n->getUniverseRadius();
    if(p.theType == Composite)
      rMax +=  2.*ParticleTable::getLargestNuclearRadius(p.theA, p.theZ);
    const double theMaxImpactParameterSquared = rMax*(rMax-theMinimumDistance);
    if(theMaxImpactParameterSquared<=0.)
      return 0.;
    const double theMaxImpactParameter = std::sqrt(theMaxImpactParameterSquared);
    return theMaxImpactParameter;
  }

  bool CoulombNonRelativistic::coulombDeviation(Particle * const p, Nucleus const * const n) const {
    // Determine the rotation angle and the new impact parameter
    ThreeVector positionTransverse = p->getTransversePosition();
    const double impactParameterSquared = positionTransverse.mag2();
    const double impactParameter = std::sqrt(impactParameterSquared);

    // Some useful variables
    const double theMinimumDistance = minimumDistance(p, n);
    // deltaTheta2 = (pi - Rutherford scattering angle)/2
    double deltaTheta2 = std::atan(2.*impactParameter/theMinimumDistance);
    if(deltaTheta2<0.)
      deltaTheta2 += Math::pi;
    const double eccentricity = 1./std::cos(deltaTheta2);

    double newImpactParameter, alpha; // Parameters that must be determined by the deviation

    const double radius = getCoulombRadius(p->getSpecies(), n);
    const double impactParameterTangentSquared = radius*(radius-theMinimumDistance);
    if(impactParameterSquared >= impactParameterTangentSquared) {
      // The particle trajectory misses the Coulomb sphere
      // In this case the new impact parameter is the minimum distance of
      // approach of the hyperbola
      assert(std::abs(1. + 2.*impactParameter*impactParameter/(radius*theMinimumDistance))>=eccentricity);
      newImpactParameter = 0.5 * theMinimumDistance * (1.+eccentricity); // the minimum distance of approach
      alpha = Math::piOverTwo - deltaTheta2; // half the Rutherford scattering angle
    } else {
      // The particle trajectory intersects the Coulomb sphere

      // Compute the entrance angle
      const double argument = -(1. + 2.*impactParameter*impactParameter/(radius*theMinimumDistance))
        / eccentricity;
      const double thetaIn = Math::twoPi - Math::arcCos(argument) - deltaTheta2;

      // Velocity angle at the entrance point
      alpha = std::atan((1+std::cos(thetaIn))
        / (std::sqrt(eccentricity*eccentricity-1.) - std::sin(thetaIn)))
        * Math::sign(theMinimumDistance);
      // New impact parameter
      newImpactParameter = radius * std::sin(thetaIn - alpha);
    }

    // Modify the impact parameter of the particle
    positionTransverse *= newImpactParameter/positionTransverse.mag();
    const ThreeVector theNewPosition = p->getLongitudinalPosition() + positionTransverse;
    p->setPosition(theNewPosition);

    // Determine the rotation axis for the incoming particle
    const ThreeVector &momentum = p->getMomentum();
    ThreeVector rotationAxis = momentum.vector(positionTransverse);
    const double axisLength = rotationAxis.mag();
    // Apply the rotation
    if(axisLength>1E-20) {
      rotationAxis /= axisLength;
      p->rotatePositionAndMomentum(alpha, rotationAxis);
    }

    return true;
  }

  double CoulombNonRelativistic::getCoulombRadius(ParticleSpecies const &p, Nucleus const * const n) const {
    if(p.theType == Composite) {
      const int Zp = p.theZ;
      const int Ap = p.theA;
      const int Zt = n->getZ();
      const int At = n->getA();
      double barr, radius = 0.;
      if(Zp==1 && Ap==2) { // d
        barr = 0.2565*Math::pow23((double)At)-0.78;
        radius = PhysicalConstants::eSquared*Zp*Zt/barr - 2.5;
      } else if(Zp==1 && Ap==3) { // t
        barr = 0.5*(0.5009*Math::pow23((double)At)-1.16);
        radius = PhysicalConstants::eSquared*Zt/barr - 0.5;
      } else if(Zp==2) { // alpha, He3
        barr = 0.5939*Math::pow23((double)At)-1.64;
        radius = PhysicalConstants::eSquared*Zp*Zt/barr - 0.5;
      } else if(Zp>2) {
        // Coulomb radius from the Shen model
        const double Ap13 = Math::pow13((double)Ap);
        const double At13 = Math::pow13((double)At);
        const double rp = 1.12*Ap13 - 0.94/Ap13;
        const double rt = 1.12*At13 - 0.94/At13;
        const double someRadius = rp+rt+3.2;
        const double theShenBarrier = PhysicalConstants::eSquared*Zp*Zt/someRadius - rt*rp/(rt+rp);
        radius = PhysicalConstants::eSquared*Zp*Zt/theShenBarrier;
      }
      if(radius<=0.) {
        radius = ParticleTable::getLargestNuclearRadius(Ap,Zp) + ParticleTable::getLargestNuclearRadius(At, Zt);
        INCL_ERROR("Negative Coulomb radius! Using the sum of nuclear radii = " << radius << '\n');
      }
      INCL_DEBUG("Coulomb radius for particle "
            << ParticleTable::getShortName(p) << " in nucleus A=" << At <<
            ", Z=" << Zt << ": " << radius << '\n');
      return radius;
    } else
      return n->getUniverseRadius();
  }

}
