/** \file G4INCLCoulombNone.cc
 * \brief Placeholder class for no Coulomb distortion.
 *
 * \date 14 February 2011
 * \author Davide Mancusi
 */

#include "G4INCLCoulombNone.hh"
#include "G4INCLIntersection.hh"

namespace G4INCL {

  ParticleEntryAvatar *CoulombNone::bringToSurface(Particle * const p, Nucleus * const n) const {
    Intersection intersection = IntersectionFactory::getEarlierTrajectoryIntersection(p->getPosition(), p->getPropagationVelocity(), n->getUniverseRadius());
    if(intersection.exists) { // If the particle enters the nucleus
      p->setPosition(intersection.position);
      return new ParticleEntryAvatar(0.0, n, p);
    } else // If the particle does NOT enter the nucleus
      return NULL;
  }

  IAvatarList CoulombNone::bringToSurface(Cluster * const c, Nucleus * const n) const {
    // The avatar list that we will return
    IAvatarList theAvatarList;

    // Loop over the particles in the cluster
    ParticleList const &projectiles = c->getParticles();
    std::list<Intersection> theIntersections;
    double theFirstEntryTime = 1E+60; // a large time
    int theFirstID = 0;
    for(ParticleIter p=projectiles.begin(), e=projectiles.end(); p!=e; ++p) {
      // Check if the particle enters the nucleus
      Intersection intersection(IntersectionFactory::getEarlierTrajectoryIntersection(
            (*p)->getPosition(),
            (*p)->getPropagationVelocity(),
            n->getUniverseRadius()));
      // Store the intersections
      theIntersections.push_back(intersection);
      if(intersection.exists) {
        // Position the particle at the entry point
        (*p)->setPosition(intersection.position);

        // Keep track of the first entering particle
        if(intersection.time < theFirstEntryTime) {
          theFirstEntryTime = intersection.time;
          theFirstID = (*p)->getID();
        }
      }
    }

    std::list<Intersection>::const_iterator intIter = theIntersections.begin();
    for(ParticleIter p=projectiles.begin(), e=projectiles.end(); p!=e; ++p, ++intIter) {

      if((*intIter).exists) {
        // If the particle enters the nucleus, generate a ParticleEntryAvatar
        // for it and add it to the list of avatars that we will return
        if((*p)->getID() == theFirstID) {
          // The first particle always enters exactly at t=0 (in order to
          // avoid negative entry times due to rounding)
          theAvatarList.push_back(new ParticleEntryAvatar(0.0, n, *p));
        } else
          theAvatarList.push_back(new ParticleEntryAvatar(intIter->time - theFirstEntryTime, n, *p));
      }

   }

    return theAvatarList;
  }

}
