#include "G4INCLCrossSections.hh"
#include "G4INCLKinematicsUtils.hh"
#include "G4INCLParticleTable.hh"
#include "G4INCLLogger.hh"
#include "G4INCLCrossSectionsINCL46.hh"
#include "G4INCLCrossSectionsMultiPions.hh"
#include "G4INCLCrossSectionsTruncatedMultiPions.hh"
#include <cassert>

namespace G4INCL {

  namespace {
    G4ThreadLocal ICrossSections *theCrossSections;
  }

  namespace CrossSections {
    double elastic(Particle const * const p1, Particle const * const p2) {
      return theCrossSections->elastic(p1,p2);
    }

    double total(Particle const * const p1, Particle const * const p2) {
      return theCrossSections->total(p1,p2);
    }

    double NDeltaToNN(Particle const * const p1, Particle const * const p2) {
      return theCrossSections->NDeltaToNN(p1,p2);
    }

    double NNToNDelta(Particle const * const p1, Particle const * const p2) {
      return theCrossSections->NNToNDelta(p1,p2);
    }

      double NNToxPiNN(const int xpi, Particle const * const p1, Particle const * const p2) {
          return theCrossSections->NNToxPiNN(xpi,p1,p2);
      }

      double piNToDelta(Particle const * const p1, Particle const * const p2) {
          return theCrossSections->piNToDelta(p1,p2);
      }

      double piNToxPiN(const int xpi, Particle const * const p1, Particle const * const p2) {
          return theCrossSections->piNToxPiN(xpi,p1,p2);
      }

    double calculateNNAngularSlope(double energyCM, int iso) {
      return theCrossSections->calculateNNAngularSlope(energyCM, iso);
    }

    double interactionDistancePiN(const double projectileKineticEnergy) {
      ThreeVector nullVector;
      ThreeVector unitVector(0., 0., 1.);

      Particle piPlusProjectile(PiPlus, unitVector, nullVector);
      piPlusProjectile.setEnergy(piPlusProjectile.getMass()+projectileKineticEnergy);
      piPlusProjectile.adjustMomentumFromEnergy();
      Particle piZeroProjectile(PiZero, unitVector, nullVector);
      piZeroProjectile.setEnergy(piZeroProjectile.getMass()+projectileKineticEnergy);
      piZeroProjectile.adjustMomentumFromEnergy();
      Particle piMinusProjectile(PiMinus, unitVector, nullVector);
      piMinusProjectile.setEnergy(piMinusProjectile.getMass()+projectileKineticEnergy);
      piMinusProjectile.adjustMomentumFromEnergy();

      Particle protonTarget(Proton, nullVector, nullVector);
      Particle neutronTarget(Neutron, nullVector, nullVector);
      const double sigmapipp = total(&piPlusProjectile, &protonTarget);
      const double sigmapipn = total(&piPlusProjectile, &neutronTarget);
      const double sigmapi0p = total(&piZeroProjectile, &protonTarget);
      const double sigmapi0n = total(&piZeroProjectile, &neutronTarget);
      const double sigmapimp = total(&piMinusProjectile, &protonTarget);
      const double sigmapimn = total(&piMinusProjectile, &neutronTarget);
      /* We compute the interaction distance from the largest of the pi-N cross
       * sections. Note that this is different from INCL4.6, which just takes the
       * average of the six, and will in general lead to a different geometrical
       * cross section.
       */
      const double largestSigma = std::max(sigmapipp, std::max(sigmapipn, std::max(sigmapi0p, std::max(sigmapi0n, std::max(sigmapimp,sigmapimn)))));
      const double interactionDistance = std::sqrt(largestSigma/Math::tenPi);

      return interactionDistance;
    }

    double interactionDistanceNN(const ParticleSpecies &aSpecies, const double kineticEnergy) {
      assert(aSpecies.theType==Proton || aSpecies.theType==Neutron || aSpecies.theType==Composite);
      assert(aSpecies.theA>0);
      ThreeVector nullVector;
      ThreeVector unitVector(0.,0.,1.);

      const double kineticEnergyPerNucleon = kineticEnergy / aSpecies.theA;

      Particle protonProjectile(Proton, unitVector, nullVector);
      protonProjectile.setEnergy(protonProjectile.getMass()+kineticEnergyPerNucleon);
      protonProjectile.adjustMomentumFromEnergy();
      Particle neutronProjectile(Neutron, unitVector, nullVector);
      neutronProjectile.setEnergy(neutronProjectile.getMass()+kineticEnergyPerNucleon);
      neutronProjectile.adjustMomentumFromEnergy();

      Particle protonTarget(Proton, nullVector, nullVector);
      Particle neutronTarget(Neutron, nullVector, nullVector);
      const double sigmapp = total(&protonProjectile, &protonTarget);
      const double sigmapn = total(&protonProjectile, &neutronTarget);
      const double sigmann = total(&neutronProjectile, &neutronTarget);
      /* We compute the interaction distance from the largest of the NN cross
       * sections. Note that this is different from INCL4.6, which just takes the
       * average of the four, and will in general lead to a different geometrical
       * cross section.
       */
      const double largestSigma = std::max(sigmapp, std::max(sigmapn, sigmann));
      const double interactionDistance = std::sqrt(largestSigma/Math::tenPi);

      return interactionDistance;
    }

    void setCrossSections(ICrossSections *c) {
      theCrossSections = c;
    }

    void deleteCrossSections() {
      delete theCrossSections;
      theCrossSections = NULL;
    }

    void initialize(Config const * const theConfig) {
      CrossSectionsType crossSections = theConfig->getCrossSectionsType();
      if(crossSections == INCL46CrossSections)
        setCrossSections(new CrossSectionsINCL46);
      else if(crossSections == MultiPionsCrossSections)
        setCrossSections(new CrossSectionsMultiPions);
      else if(crossSections == TruncatedMultiPionsCrossSections) {
        const int nMaxPi = theConfig->getMaxNumberMultipions();
        if(nMaxPi>0)
          setCrossSections(new CrossSectionsTruncatedMultiPions(nMaxPi));
        else {
          INCL_WARN("Truncated multipion cross sections were requested, but the specified maximum\n"
                    << "number of pions is <=0. Falling back to standard multipion cross-sections.\n");
          setCrossSections(new CrossSectionsMultiPions);
        }
      }
    }
  }
}
