/** \file G4INCLDeuteronDensity.cc
 * \brief Deuteron density in r and p according to the Paris potential.
 *
 * \date 6 March 2012
 * \author Davide Mancusi
 */

#include "G4INCLDeuteronDensity.hh"
#include "G4INCLGlobals.hh"
#include <cassert>
#include <algorithm>

namespace G4INCL {

  namespace DeuteronDensity {

    namespace {

      const int coeffTableSize = 13;

      /// \brief Coefficients for the deuteron wave function
      const double coeff1[coeffTableSize] = {
        0.88688076e+0,
        -0.34717093e+0,
        -0.30502380e+1,
        0.56207766e+2,
        -0.74957334e+3,
        0.53365279e+4,
        -0.22706863e+5,
        0.60434469e+5,
        -0.10292058e+6,
        0.11223357e+6,
        -0.75925226e+5,
        0.29059715e+5,
        -0.48157368e+4
      };

      /// \brief Coefficients for the deuteron wave function
      const double coeff2[coeffTableSize] = {
        0.23135193e-1,
        -0.85604572e+0,
        0.56068193e+1,
        -0.69462922e+2,
        0.41631118e+3,
        -0.12546621e+4,
        0.12387830e+4,
        0.33739172e+4,
        -0.13041151e+5,
        0.19512524e+5,
        -0.15634324e+5,
        0.66231089e+4,
        -0.11698185e+4
      };

      /// \brief Normalisation coefficient for the r-space deuteron wave function
      const double normalisationR = std::sqrt(32. * Math::pi) * 0.28212;

      /// \brief Normalisation coefficient for the p-space deuteron wave function
      const double normalisationP = normalisationR / (std::sqrt(4. * Math::pi) * std::pow(PhysicalConstants::hc,1.5));

      /// \brief Mysterious coefficient that appears in the wavefunctions
      const double al = 0.23162461;

    }

    double densityR(const double r) {
      const double sWave = wavefunctionR(0, r);
      const double dWave = wavefunctionR(2, r);
      return r*r*(sWave*sWave + dWave*dWave);
    }

    double derivDensityR(const double r) {
      const double sWave = wavefunctionR(0, r);
      const double dWave = wavefunctionR(2, r);
      const double sWaveDeriv = derivWavefunctionR(0, r);
      const double dWaveDeriv = derivWavefunctionR(2, r);
      return (sWave*sWaveDeriv + dWave*dWaveDeriv) / Math::twoPi;
    }

    double densityP(const double p) {
      const double sWave = wavefunctionP(0, p);
      const double dWave = wavefunctionP(2, p);
      return p*p*(sWave*sWave + dWave*dWave);
    }

    double wavefunctionR(const int l, const double theR) {
      assert(l==0 || l==2); // only s- and d-waves in a deuteron
      const double r = 2. * std::max(theR, 1.e-4);

      double result = 0.;
      double fmr;

      for(int i=0; i<coeffTableSize; ++i) {
        fmr = r * (al+i);
        if(l==0) { // s-wave
          result += coeff1[i] * std::exp(-fmr);
        } else { // d-wave
          result += coeff2[i] * std::exp(-fmr) * (1.+3./fmr+3./(fmr*fmr));
        }
      }

      result *= normalisationR/r;
      return result;
    }

    double derivWavefunctionR(const int l, const double theR) {
      assert(l==0 || l==2); // only s- and d-waves in a deuteron
      const double r = 2. * std::max(theR, 1.e-4);

      double result = 0.;
      double fmr;

      for(int i=0; i<coeffTableSize; ++i) {
        fmr = r * (al+i);
        if(l==0) { // s-wave
          result += coeff1[i] * std::exp(-fmr) * (fmr + 1.);
        } else { // d-wave
          result += coeff2[i] * std::exp(-fmr) * (fmr + 4. + 9./fmr + 9./(fmr*fmr));
        }
      }

      result *= -normalisationR/(r*r);
      return result;
    }

    double wavefunctionP(const int l, const double theQ) {
      assert(l==0 || l==2); // only s- and d-waves in a deuteron
      const double q = theQ / PhysicalConstants::hc;
      const double q2 = q*q;
      double result=0.;
      double fmq, alPlusI;
      for(int i=0; i<coeffTableSize; ++i) {
        alPlusI = al+i;
        fmq = q2 + alPlusI*alPlusI;
        if(l==0) { // s-wave
          result += coeff1[i] / fmq;
        } else { // d-wave
          result += coeff2[i] / fmq;
        }
      }

      result *= normalisationP;
      return result;
    }

  }

}
