/** \file G4INCLINuclearPotential.cc
 * \brief Abstract interface to the nuclear potential.
 *
 * NuclearPotential-like classes should provide access to the value of the
 * potential of a particle in a particular context. For example, an instance of
 * a NuclearPotential class should be associated to every nucleus.
 *
 * \date 31 March 2011
 * \author Davide Mancusi
 */

#include "G4INCLINuclearPotential.hh"
#include "G4INCLNuclearPotentialEnergyIsospinSmooth.hh"
#include "G4INCLNuclearPotentialEnergyIsospin.hh"
#include "G4INCLNuclearPotentialIsospin.hh"
#include "G4INCLNuclearPotentialConstant.hh"

namespace G4INCL {

  namespace NuclearPotential {

    const double INuclearPotential::vPionDefault = 30.6; // MeV

    namespace {

      G4ThreadLocal std::map<long,INuclearPotential const *> *nuclearPotentialCache = NULL;

    }

    INuclearPotential const *createPotential(const PotentialType type, const int theA, const int theZ, const bool pionPotential) {
      if(!nuclearPotentialCache)
        nuclearPotentialCache = new std::map<long,INuclearPotential const *>;

      const long nuclideID = (pionPotential ? 1 : -1) * (1000*theZ + theA + 1000000*type); // MCNP-style nuclide IDs
      const std::map<long,INuclearPotential const *>::const_iterator mapEntry = nuclearPotentialCache->find(nuclideID);
      if(mapEntry == nuclearPotentialCache->end()) {
        INuclearPotential const *thePotential = NULL;
        switch(type) {
          case IsospinEnergySmoothPotential:
            thePotential = new NuclearPotentialEnergyIsospinSmooth(theA, theZ, pionPotential);
            break;
          case IsospinEnergyPotential:
            thePotential = new NuclearPotentialEnergyIsospin(theA, theZ, pionPotential);
            break;
          case IsospinPotential:
            thePotential = new NuclearPotentialIsospin(theA, theZ, pionPotential);
            break;
          case ConstantPotential:
            thePotential = new NuclearPotentialConstant(theA, theZ, pionPotential);
            break;
          default:
            INCL_FATAL("Unrecognized potential type at Nucleus creation." << '\n');
            break;
        }
        (*nuclearPotentialCache)[nuclideID] = thePotential;
        return thePotential;
      } else {
        return mapEntry->second;
      }
    }

    void clearCache() {
      if(nuclearPotentialCache) {
        for(std::map<long,INuclearPotential const *>::const_iterator i = nuclearPotentialCache->begin(), e=nuclearPotentialCache->end(); i!=e; ++i)
          delete i->second;
        nuclearPotentialCache->clear();
        delete nuclearPotentialCache;
        nuclearPotentialCache = NULL;
      }
    }

  } // namespace NuclearPotential

} // namespace G4INCL

