#include "G4INCLKinematicsUtils.hh"
#include "G4INCLParticleTable.hh"

namespace G4INCL {

  namespace KinematicsUtils {

  void transformToLocalEnergyFrame(Nucleus const * const n, Particle * const p) {
    const double localEnergy = getLocalEnergy(n, p);
    const double localTotalEnergy = p->getEnergy() - localEnergy;
    p->setEnergy(localTotalEnergy);
    p->adjustMomentumFromEnergy();
  }

  double getLocalEnergy(Nucleus const * const n, Particle * const p) {
    assert(!p->isPion()); // No local energy for pions

    double vloc = 0.0;
    const double r = p->getPosition().mag();
    const double mass = p->getMass();

    // Local energy is constant outside the surface
    if(r > n->getUniverseRadius()) {
      INCL_WARN("Tried to evaluate local energy for a particle outside the maximum radius."
            << '\n' << p->print() << '\n'
            << "Maximum radius = " << n->getDensity()->getMaximumRadius() << '\n'
            << "Universe radius = " << n->getUniverseRadius() << '\n');
      return 0.0;
    }

    double pfl0 = 0.0;
    const ParticleType t = p->getType();
    const double kinE = p->getKineticEnergy();
    if(kinE <= n->getPotential()->getFermiEnergy(t)) {
      pfl0 = n->getPotential()->getFermiMomentum(p);
    } else {
      const double tf0 = p->getPotentialEnergy() - n->getPotential()->getSeparationEnergy(p);
      if(tf0<0.0) return 0.0;
      pfl0 = std::sqrt(tf0*(tf0 + 2.0*mass));
    }
    const double pReflection = p->getReflectionMomentum()/pfl0;
    const double reflectionRadius = n->getDensity()->getMaxRFromP(p->getType(), pReflection);
    const double pNominal = p->getMomentum().mag()/pfl0;
    const double nominalReflectionRadius = n->getDensity()->getMaxRFromP(p->getType(), pNominal);
    const double pl = pfl0*n->getDensity()->getMinPFromR(t, r*nominalReflectionRadius/reflectionRadius);
    vloc = std::sqrt(pl*pl + mass*mass) - mass;

    return vloc;
  }

  ThreeVector makeBoostVector(Particle const * const p1, Particle const * const p2){
    const double totalEnergy = p1->getEnergy() + p2->getEnergy();
    return ((p1->getMomentum() + p2->getMomentum())/totalEnergy);
  }

  double totalEnergyInCM(Particle const * const p1, Particle const * const p2){
    return std::sqrt(squareTotalEnergyInCM(p1,p2));
  }

  double squareTotalEnergyInCM(Particle const * const p1, Particle const * const p2) {
    double beta2 = makeBoostVector(p1, p2).mag2();
    if(beta2 > 1.0) {
      INCL_ERROR("squareTotalEnergyInCM: beta2 == " << beta2 << " > 1.0" << '\n');
      beta2 = 0.0;
    }
    return (1.0 - beta2)*std::pow(p1->getEnergy() + p2->getEnergy(), 2);
  }

  double momentumInCM(Particle const * const p1, Particle const * const p2) {
    const double m1sq = std::pow(p1->getMass(),2);
    const double m2sq = std::pow(p2->getMass(),2);
    const double z = p1->getEnergy()*p2->getEnergy() - p1->getMomentum().dot(p2->getMomentum());
    double pcm2 = (z*z-m1sq*m2sq)/(2*z+m1sq+m2sq);
    if(pcm2 < 0.0) {
      INCL_ERROR("momentumInCM: pcm2 == " << pcm2 << " < 0.0" << '\n');
      pcm2 = 0.0;
    }
    return std::sqrt(pcm2);
  }

  double momentumInCM(const double E, const double M1, const double M2) {
    return 0.5*std::sqrt((E*E - std::pow(M1 + M2, 2))
			 *(E*E - std::pow(M1 - M2, 2)))/E;
  }

  double momentumInLab(const double s, const double m1, const double m2) {
    const double m1sq = m1*m1;
    const double m2sq = m2*m2;
    double plab2 = (s*s-2*s*(m1sq+m2sq)+(m1sq-m2sq)*(m1sq-m2sq))/(4*m2sq);
    if(plab2 < 0.0) {
      INCL_ERROR("momentumInLab: plab2 == " << plab2 << " < 0.0; m1sq == " << m1sq << "; m2sq == " << m2sq << "; s == " << s << '\n');
      plab2 = 0.0;
    }
    return std::sqrt(plab2);
  }

  double momentumInLab(Particle const * const p1, Particle const * const p2) {
    const double m1 = p1->getMass();
    const double m2 = p2->getMass();
    const double s = squareTotalEnergyInCM(p1, p2);
    return momentumInLab(s, m1, m2);
  }

  double sumTotalEnergies(const ParticleList &pl) {
    double E = 0.0;
    for(ParticleIter i=pl.begin(), e=pl.end(); i!=e; ++i) {
      E += (*i)->getEnergy();
    }
    return E;
  }

  ThreeVector sumMomenta(const ParticleList &pl) {
    ThreeVector p(0.0, 0.0, 0.0);
    for(ParticleIter i=pl.begin(), e=pl.end(); i!=e; ++i) {
      p += (*i)->getMomentum();
    }
    return p;
  }

  double energy(const ThreeVector &p, const double m) {
    return std::sqrt(p.mag2() + m*m);
  }

  double invariantMass(const double E, const ThreeVector & p) {
    return std::sqrt(squareInvariantMass(E, p));
  }

  double squareInvariantMass(const double E, const ThreeVector & p) {
    return E*E - p.mag2();
  }

  double gammaFromKineticEnergy(const ParticleSpecies &p, const double EKin) {
    double mass;
    if(p.theType==Composite)
      mass = ParticleTable::getTableMass(p.theA, p.theZ);
    else
      mass = ParticleTable::getTableParticleMass(p.theType);
    return (1.+EKin/mass);
  }

  }

}
