/** \file G4INCLNuclearPotentialEnergyIsospinSmooth.cc
 * \brief Isospin- and energy-dependent nuclear potential.
 *
 * Provides an isospin- and energy-dependent nuclear potential. The cusp at 200
 * MeV is replaced by a smooth exponential.
 *
 * \date 16 February 2011
 * \author Davide Mancusi
 */

#include "G4INCLNuclearPotentialEnergyIsospinSmooth.hh"
#include "G4INCLParticleTable.hh"
#include "G4INCLGlobals.hh"

namespace G4INCL {

  namespace NuclearPotential {

    const double NuclearPotentialEnergyIsospinSmooth::alpha= 0.223;
    const double NuclearPotentialEnergyIsospinSmooth::deltaE= 25.;

    // Constructors
    NuclearPotentialEnergyIsospinSmooth::NuclearPotentialEnergyIsospinSmooth(const int A, const int Z, const bool aPionPotential)
      : NuclearPotentialIsospin(A,Z,aPionPotential)
    {}

    // Destructor
    NuclearPotentialEnergyIsospinSmooth::~NuclearPotentialEnergyIsospinSmooth() {}

    double NuclearPotentialEnergyIsospinSmooth::computePotentialEnergy(const Particle *particle) const {

      const double v0 = NuclearPotentialIsospin::computePotentialEnergy(particle);

      if(particle->isNucleon()) {
        const double t = particle->getKineticEnergy();
        const double tf = getFermiEnergy(particle);
        // Constant potential for T<Tf
        if(t < tf)
          return v0;

        // Linear function for Tf<T<T0, exponential function for T>T0
        const double t0 = tf + v0*(1.-alpha)/alpha - deltaE; // deltaE before the linear potential vanishes
        double v;
        if(t<t0) {
          v = v0 - alpha*(t-tf)/(1.-alpha);
        } else {
          const double v_at_t0 = v0 - alpha*(t0-tf)/(1.-alpha);
          const double kappa = alpha / (v_at_t0 * (1.-alpha));
          v = v_at_t0 * std::exp(kappa * (t0-t));
        }
        return (v>0.0) ? v : 0.0;
      } else
        return v0;
    }

  }
}

