#include "G4INCLParticleEntryFSIChannel.hh"
#include "G4INCLRootFinder.hh"
#include "G4INCLIntersection.hh"
#include <algorithm>

namespace G4INCL {

  ParticleEntryFSIChannel::ParticleEntryFSIChannel(Nucleus *n, Particle *p)
    :theNucleus(n), theParticle(p)
  {}

  ParticleEntryFSIChannel::~ParticleEntryFSIChannel()
  {}

  void ParticleEntryFSIChannel::fillFinalState(FinalState *fs) {
   /*
    * Corrections to the energy of the entering nucleon
    *
    * In particle-nucleus reactions, the goal of this correction is to satisfy
    * energy conservation in particle-nucleus reactions using real particle
    * and nuclear masses.
    *
    * Possible choices for the correction (or, equivalently, for the QP
    * excitation energy):
    *
    * 1. the correction is 0. (same as in particle-nucleus);
    * 2. the correction is the separation energy of the entering nucleon in
    *    the current QP;
    * 3. the QP excitation energy is given by A. Boudard's algorithm, as
    *    implemented in INCL4.2-HI/Geant4.
    * 4. the QP excitation energy vanishes.
    */
    const double energyBefore = theParticle->getEnergy() ;
    bool success = particleInside(0); // theQValueCorrection=0 same as in particle-nucleus
    fs->addEnteringParticle(theParticle);
    fs->setTotalEnergyBeforeInteraction(energyBefore);
  }

  bool ParticleEntryFSIChannel::particleInside(const double theQValueCorrection) {

    theParticle->setINCLMass(); // Will automatically put the particle on shell

    // Add the nuclear potential to the kinetic energy when entering the
    // nucleus

    class IncomingEFunctor : public RootFunctor {
      public:
        IncomingEFunctor(Particle * const p, Nucleus const * const n, const double correction) :
          RootFunctor(0., 1E6),
          theParticle(p),
          thePotential(n->getPotential()),
          theEnergy(theParticle->getEnergy()),
          theMass(theParticle->getMass()),
          theQValueCorrection(correction),
          refraction(n->getStore()->getConfig()->getRefraction()),
          theMomentumDirection(theParticle->getMomentum())
          {
            if(refraction) {
              const ThreeVector &position = theParticle->getPosition();
              const double r2 = position.mag2();
              if(r2>0.)
                normal = - position / std::sqrt(r2);
              double cosIncidenceAngle = theParticle->getCosRPAngle();
              if(cosIncidenceAngle < -1.)
                sinIncidenceAnglePOut = 0.;
              else
                sinIncidenceAnglePOut = theMomentumDirection.mag()*std::sqrt(1.-cosIncidenceAngle*cosIncidenceAngle);
            } else {
              sinIncidenceAnglePOut = 0.;
            }
          }
        ~IncomingEFunctor() {}
        double operator()(const double v) const {
          double energyInside = std::max(theMass, theEnergy + v - theQValueCorrection);
          theParticle->setEnergy(energyInside);
          theParticle->setPotentialEnergy(v);
          if(refraction) {
            // Compute the new direction of the particle momentum
            const double pIn = std::sqrt(energyInside*energyInside-theMass*theMass);
            const double sinRefractionAngle = sinIncidenceAnglePOut/pIn;
            const double cosRefractionAngle = (sinRefractionAngle>1.) ? 0. : std::sqrt(1.-sinRefractionAngle*sinRefractionAngle);
            const ThreeVector momentumInside = theMomentumDirection - normal * normal.dot(theMomentumDirection) + normal * (pIn * cosRefractionAngle);
            theParticle->setMomentum(momentumInside);
          } else {
            theParticle->setMomentum(theMomentumDirection); // keep the same direction
          }
          // Scale the particle momentum
          theParticle->adjustMomentumFromEnergy();
          return v - thePotential->computePotentialEnergy(theParticle);
        }
        void cleanUp(const bool /*success*/) const {}
      private:
        Particle *theParticle;
        NuclearPotential::INuclearPotential const *thePotential;
        const double theEnergy;
        const double theMass;
        const double theQValueCorrection;
        const bool refraction;
        const ThreeVector theMomentumDirection;
        ThreeVector normal;
        double sinIncidenceAnglePOut;
    } theIncomingEFunctor(theParticle,theNucleus,theQValueCorrection);

    double v = theNucleus->getPotential()->computePotentialEnergy(theParticle);
    if(theParticle->getKineticEnergy()+v-theQValueCorrection<0.) { // Particle entering below 0. Die gracefully
      INCL_DEBUG("Particle " << theParticle->getID() << " is trying to enter below 0" << '\n');
      return false;
    }
    const RootFinder::Solution theSolution = RootFinder::solve(&theIncomingEFunctor, v);
    if(theSolution.success) { // Apply the solution
      theIncomingEFunctor(theSolution.x);
      INCL_DEBUG("Particle successfully inside the nucleus:\n" << theParticle->print() << '\n');
    } else {
      INCL_WARN("Couldn't compute the potential for secondary, root-finding algorithm failed." << '\n');
    }
    return theSolution.success;
  }

}

