#include "G4INCLPauliStandard.hh"
#include "G4INCLPauliBlocking.hh"
#include "G4INCLGlobals.hh"
#include "G4INCLRandom.hh"

namespace G4INCL {


  PauliStandard::PauliStandard() :
    cellSize(std::pow(2.38*4.5*Math::pi,1./6.)*std::sqrt(PhysicalConstants::hc))
  {
    INCL_DEBUG("Initialising PauliStandard. cellSize=" << cellSize << '\n');
  }

  PauliStandard::~PauliStandard() {}

  bool PauliStandard::isBlocked(ParticleList const &pL, Nucleus const * const n) {
    for(ParticleIter p=pL.begin(), e=pL.end(); p!=e; ++p) {
      if( !(*p)->isNucleon() ) continue;
      if(getBlockingProbability(*p, n) > Random::shoot()) return true;
    }
    return false;
  }

  double PauliStandard::getBlockingProbability(Particle const * const particle, Nucleus const * const nucleus) const {
    const double r0 = ParticleTable::getNuclearRadius(particle->getType(), nucleus->getA(), nucleus->getZ());
    const double pFermi = nucleus->getPotential()->getFermiMomentum(particle);

    const double pbl = cellSize * std::sqrt(pFermi/r0);
    const double rbl = pbl * r0/pFermi;
    const double maxVolR = rbl;
    const double maxVolP = pbl;
    double vol = std::pow(4.*Math::pi/3.0, 2)
      * std::pow(maxVolR*maxVolP/(Math::twoPi*PhysicalConstants::hc), 3);

    const double rdeq = nucleus->getUniverseRadius();
    const double rs = particle->getPosition().mag();

    if(rs - maxVolR > rdeq) {
      return 0.0;
    }

    if(rs + maxVolR > rdeq) {
      vol = vol * 0.5 * (rdeq - rs + maxVolR) / maxVolR;
    }

    // Get the list of particles that are currently inside the
    // nucleus.
    ParticleList const &particles = nucleus->getStore()->getParticles();

    int nl = 0;
    for(ParticleIter it=particles.begin(), e=particles.end(); it!=e; ++it) {
      // Skip comparing with the same particle
      if( (*it)->getID() == particle->getID() ) continue;

      if((*it)->getType() == particle->getType()) {
        const ThreeVector dx2v = particle->getPosition() - (*it)->getPosition();
        const double dx2 = dx2v.mag2();
        if(dx2 > maxVolR * maxVolR) continue;

        const ThreeVector dp2v = particle->getMomentum() - (*it)->getMomentum();
        const double dp2 = dp2v.mag2();
        if(dp2 > maxVolP * maxVolP) continue;

        nl++;
      }
    }
    const double blockingProbability = ((double) nl) / vol / 2.0;

    if(blockingProbability > 1.0) return 1.0;
    else if(blockingProbability < 0.0) return 0.0;
    else return blockingProbability;
  }
}
