#include "G4INCLPhaseSpaceGenerator.hh"
#include "G4INCLPhaseSpaceKopylov.hh"
#include "G4INCLPhaseSpaceRauboldLynch.hh"

namespace G4INCL {

  namespace {
    G4ThreadLocal IPhaseSpaceGenerator *thePhaseSpaceGenerator;

    G4ThreadLocal Particle *biasMe;

    /** \brief Actually perform the biasing
     *
     * \param particles list of particles to bias
     * \param pInVec momentum of the particle to be biased before the collision
     * \param slope the parameter \f$B\f$ in \f$\exp(B\cdot t)\f$
     */
    void bias(ParticleList &particles, const ThreeVector &pInVec, const double slope) {
      const double pIn = pInVec.mag();
      const ThreeVector collisionAxis = pInVec/pIn;
      const ThreeVector pMomVec = biasMe->getMomentum();
      const double pMom = pMomVec.mag();
      const double pMomCosAng = pMomVec.dot(collisionAxis)/pMom;
      const double pMomAng = Math::arcCos(pMomCosAng);

      // compute the target angle for the biasing
      // it is drawn from a exp(Bt) distribution
      const double cosAngSlope = 2e-6 * slope * pIn * pMom;
      const double cosAng = 1. + std::log(1. - Random::shoot()*(1.-std::exp(-2.*cosAngSlope)))/cosAngSlope;
      const double ang = Math::arcCos(cosAng);

      // compute the rotation angle
      const double rotationAngle = ang - pMomAng;

      // generate the rotation axis; it is perpendicular to collisionAxis and
      // pMomVec
      ThreeVector rotationAxis;
      if(pMomAng>1E-10) {
        rotationAxis = collisionAxis.vector(pMomVec);
        const double axisLength = rotationAxis.mag();
        const double oneOverLength = 1./axisLength;
        rotationAxis *= oneOverLength;
      } else {
        // need to jump through some hoops if collisionAxis is nearly aligned
        // with pMomVec
        rotationAxis = collisionAxis.anyOrthogonal();
      }

      // apply the rotation
      particles.rotateMomentum(rotationAngle, rotationAxis);
    }

  }

  namespace PhaseSpaceGenerator {
    void generate(const double sqrtS, ParticleList &particles) {
      return thePhaseSpaceGenerator->generate(sqrtS, particles);
    }

    void generateBiased(const double sqrtS, ParticleList &particles, const size_t index, const double slope) {
      assert(index<particles.size());
      // store the incoming momentum of particle[index]; it will be used to
      // compute t when biasing
      biasMe = particles[index];
      const ThreeVector pInVec = biasMe->getMomentum();
      generate(sqrtS, particles);
      bias(particles, pInVec, slope);
    }

    void setPhaseSpaceGenerator(IPhaseSpaceGenerator *g) {
      thePhaseSpaceGenerator = g;
    }

    IPhaseSpaceGenerator *getPhaseSpaceGenerator() {
      return thePhaseSpaceGenerator;
    }

    void deletePhaseSpaceGenerator() {
      delete thePhaseSpaceGenerator;
      thePhaseSpaceGenerator = NULL;
    }

    void initialize(Config const * const theConfig) {
      PhaseSpaceGeneratorType psg = theConfig->getPhaseSpaceGeneratorType();
      if(psg==RauboldLynchType)
        setPhaseSpaceGenerator(new PhaseSpaceRauboldLynch);
      else if(psg==KopylovType)
        setPhaseSpaceGenerator(new PhaseSpaceKopylov);
      else
        setPhaseSpaceGenerator(NULL);
    }
  }
}
