#include "G4INCLPhaseSpaceKopylov.hh"
#include "G4INCLRandom.hh"
#include "G4INCLKinematicsUtils.hh"
#include <algorithm>
#include <numeric>
#include <functional>

namespace G4INCL {

  double PhaseSpaceKopylov::betaKopylov(int K) const {
    int N = 3*K - 5;
    double xN = double(N);
    double Fmax = std::sqrt(std::pow(xN/(xN+1.),N)/(xN+1.));

    double F, chi;
    unsigned long loopCounter = 0;
    const unsigned long maxLoopCounter = 10000000;
    do {
      chi = Random::shoot();
      F = std::sqrt(std::pow(chi,N)*(1.-chi));
      ++loopCounter;
    } while (loopCounter<maxLoopCounter && Fmax*Random::shoot() > F); /* Loop checking, 10.07.2015, D.Mancusi */
    return chi;
  }

  void PhaseSpaceKopylov::generate(const double sqrtS, ParticleList &particles) {

    boostV.setX(0.0);
    boostV.setY(0.0);
    boostV.setZ(0.0);

    const size_t N = particles.size();
    masses.resize(N);
    sumMasses.resize(N);
    std::transform(particles.begin(), particles.end(), masses.begin(), std::mem_fun(&Particle::getMass));
    std::partial_sum(masses.begin(), masses.end(), sumMasses.begin());

    double PFragMagCM = 0.0;
    double T = sqrtS-sumMasses.back();
    assert(T>-1.e-5);
    if(T<0.)
      T=0.;

    // The first particle in the list will pick up all the recoil
    Particle *restParticle = particles.front();
    restParticle->setMass(sqrtS);
    restParticle->adjustEnergyFromMomentum();

    int k=N-1;
    for (ParticleList::reverse_iterator p=particles.rbegin(); k>0; ++p, --k) {
      const double mu = sumMasses[k-1];
      T *= (k>1) ? betaKopylov(k) : 0.;

      const double restMass = mu + T;

      PFragMagCM = KinematicsUtils::momentumInCM(restParticle->getMass(), masses[k], restMass);
      PFragCM = Random::normVector(PFragMagCM);
      (*p)->setMomentum(PFragCM);
      (*p)->adjustEnergyFromMomentum();
      restParticle->setMass(restMass);
      restParticle->setMomentum(-PFragCM);
      restParticle->adjustEnergyFromMomentum();

      (*p)->boost(boostV);
      restParticle->boost(boostV);

      boostV = -restParticle->boostVector();
    }
    restParticle->setMass(masses[0]);
    restParticle->adjustEnergyFromMomentum();
  }

}
