/** \file G4INCLRecombinationChannel.cc
 * \brief Delta-nucleon recombination channel.
 *
 *  \date 25 March 2011
 * \author Davide Mancusi
 */

#include "G4INCLRecombinationChannel.hh"
#include "G4INCLRandom.hh"
#include "G4INCLKinematicsUtils.hh"
#include "G4INCLParticleTable.hh"
#include "G4INCLBinaryCollisionAvatar.hh"
#include "G4INCLGlobals.hh"

#include <cassert>

namespace G4INCL {

  RecombinationChannel::RecombinationChannel(Particle *p1, Particle *p2)
  {
    if(p1->isDelta()) {
      assert(p2->isNucleon());
      theDelta = p1;
      theNucleon = p2;
    } else {
      assert(p1->isNucleon());
      theDelta = p2;
      theNucleon = p1;
    }
  }

  RecombinationChannel::~RecombinationChannel()
  {
  }

  void RecombinationChannel::fillFinalState(FinalState *fs)
  {
    // Compute the total available energy in the CM
    const double sqrts = KinematicsUtils::totalEnergyInCM(theDelta, theNucleon);

    // Assign the types of the final-state particles
    switch(theDelta->getType()) {
      case DeltaPlusPlus:
        assert(theNucleon->getType()!=Proton);
        theDelta->setType(Proton);
        theNucleon->setType(Proton);
        break;
      case DeltaPlus:
        theDelta->setType(Proton);
        break;
      case DeltaZero:
        theDelta->setType(Neutron);
        break;
      case DeltaMinus:
        assert(theNucleon->getType()!=Neutron);
        theDelta->setType(Neutron);
        theNucleon->setType(Neutron);
        break;
      default:
        INCL_ERROR("Unknown particle type in RecombinationChannel" << '\n');
        break;
    }

    // Calculate the momenta of the nucleons in the final state
    const double pCM = KinematicsUtils::momentumInCM(sqrts, theDelta->getMass(), theNucleon->getMass());

    // The angular distribution of final-state nucleons is isotropic
    ThreeVector momentum = Random::normVector(pCM);

    // Assign the momenta
    theDelta->setMomentum(momentum);
    theNucleon->setMomentum(-momentum);

    // Update the kinetic energies
    theDelta->adjustEnergyFromMomentum();
    theNucleon->adjustEnergyFromMomentum();

    // Create the final state
    fs->addModifiedParticle(theDelta);
    fs->addModifiedParticle(theNucleon);

  }

}
