#include "G4INCLReflectionChannel.hh"
#include "G4INCLFinalState.hh"
#include "G4INCLRandom.hh"
#include "G4INCLINuclearPotential.hh"

#include <cmath>

namespace G4INCL {
  const double ReflectionChannel::sinMinReflectionAngleSquaredOverFour = std::pow(std::sin(2.*Math::pi/200.),2.);
  const double ReflectionChannel::positionScalingFactor = 0.99;

  ReflectionChannel::ReflectionChannel(Nucleus *n, Particle *p)
    :theNucleus(n),theParticle(p)
  {
  }

  ReflectionChannel::~ReflectionChannel()
  {
  }

  void ReflectionChannel::fillFinalState(FinalState *fs) {
    fs->setTotalEnergyBeforeInteraction(theParticle->getEnergy() - theParticle->getPotentialEnergy());

    const ThreeVector &oldMomentum = theParticle->getMomentum();
    const ThreeVector thePosition = theParticle->getPosition();
    double pspr = thePosition.dot(oldMomentum);
    if(pspr>=0) { // This means that the particle is trying to leave; perform a reflection
      const double x2cour = thePosition.mag2();
      const ThreeVector newMomentum = oldMomentum - (thePosition * (2.0 * pspr/x2cour));
      const double deltaP2 = (newMomentum-oldMomentum).mag2();
      theParticle->setMomentum(newMomentum);
      const double minDeltaP2 = sinMinReflectionAngleSquaredOverFour * newMomentum.mag2();
      if(deltaP2 < minDeltaP2) { // Avoid extremely small reflection angles
        theParticle->setPosition(thePosition * positionScalingFactor);
        INCL_DEBUG("Reflection angle for particle " << theParticle->getID() << " was too tangential: " << '\n'
            << "  " << deltaP2 << "=deltaP2<minDeltaP2=" << minDeltaP2 << '\n'
            << "  Resetting the particle position to ("
            << thePosition.getX() << ", "
            << thePosition.getY() << ", "
            << thePosition.getZ() << ")" << '\n');
      }
      theNucleus->updatePotentialEnergy(theParticle);
    } else { // The particle momentum is already directed towards the inside of the nucleus; do nothing
      // ...but make sure this only happened because of the frozen propagation
      assert(theParticle->getPosition().dot(theParticle->getPropagationVelocity())>0.);
    }

    theParticle->thawPropagation();
    fs->addModifiedParticle(theParticle);
  }
}

