#include "G4INCLTransmissionChannel.hh"

namespace G4INCL {

  TransmissionChannel::TransmissionChannel(Nucleus * const nucleus, Particle * const particle)
    : theNucleus(nucleus), theParticle(particle),
    refraction(false),
    pOutMag(0.),
    kineticEnergyOutside(initializeKineticEnergyOutside()),
    cosRefractionAngle(1.)
  {}

  TransmissionChannel::TransmissionChannel(Nucleus * const nucleus, Particle * const particle, const double TOut)
    : theNucleus(nucleus), theParticle(particle),
    refraction(false),
    pOutMag(0.),
    kineticEnergyOutside(TOut),
    cosRefractionAngle(1.)
  {}

  TransmissionChannel::TransmissionChannel(Nucleus * const nucleus, Particle * const particle, const double kOut, const double cosR)
    : theNucleus(nucleus), theParticle(particle),
    refraction(true),
    pOutMag(kOut),
    kineticEnergyOutside(initializeKineticEnergyOutside()),
    cosRefractionAngle(cosR)
  {}

  TransmissionChannel::~TransmissionChannel() {}

  double TransmissionChannel::initializeKineticEnergyOutside() {
    // The particle energy outside the nucleus. Subtract the nuclear
    // potential from the kinetic energy when leaving the nucleus
    double TOut = theParticle->getEnergy()
      - theParticle->getPotentialEnergy()
      - theParticle->getMass();

    // Correction for real masses
    const int AParent = theNucleus->getA();
    const int ZParent = theNucleus->getZ();
    const double theQValueCorrection = theParticle->getEmissionQValueCorrection(AParent,ZParent);
    TOut += theQValueCorrection;
    return TOut;
  }

  void TransmissionChannel::particleLeaves() {

    // Use the table mass in the outside world
    theParticle->setTableMass();
    theParticle->setPotentialEnergy(0.);

    if(refraction) {
      // Change the momentum direction
      // The magnitude of the particle momentum outside the nucleus will be
      // fixed by the kineticEnergyOutside variable. This is done in order to
      // avoid numerical inaccuracies.
      const ThreeVector &position = theParticle->getPosition();
      const double r2 = position.mag2();
      ThreeVector normal;
      if(r2>0.)
        normal = position / std::sqrt(r2);

      const ThreeVector &momentum = theParticle->getMomentum();

      const ThreeVector pOut = normal * (pOutMag * cosRefractionAngle) + momentum - normal * normal.dot(momentum);
      assert(std::fabs(pOut.mag()-pOutMag)<1.e-5);

      theParticle->setMomentum(pOut);
    }
    // Scaling factor for the particle momentum
    theParticle->setEnergy(kineticEnergyOutside + theParticle->getMass());
    theParticle->adjustMomentumFromEnergy();
  }

  void TransmissionChannel::fillFinalState(FinalState *fs) {
    double initialEnergy = 0.0;
    initialEnergy = theParticle->getEnergy() - theParticle->getPotentialEnergy();

    // Correction for real masses
    const int AParent = theNucleus->getA();
    const int ZParent = theNucleus->getZ();
    initialEnergy += theParticle->getTableMass() - theParticle->getMass()
      + theParticle->getEmissionQValueCorrection(AParent,ZParent);

    particleLeaves();

    fs->setTotalEnergyBeforeInteraction(initialEnergy);
    fs->addOutgoingParticle(theParticle); // We write the particle down as outgoing
  }
}
