//============================================================================
// Name        : INCLCascade.cpp
// Author      : Pekka Kaitaniemi
// Version     :
// Copyright   : Copyright (c) 2009 Alain Boudard, Joseph Cugnon, Sylvie Leray, Pekka Kaitaniemi
// Description : INCL++ intra-nuclear cascade simulation
//============================================================================

#include <iostream>
#include <iomanip>
#include <string>
#include <sstream>
#include <map>

#include "G4INCLCascade.hh"

#include "G4INCLParticle.hh"
#include "G4INCLIAvatar.hh"
#include "G4INCLIPropagationModel.hh"
#include "G4INCLNucleus.hh"
#include "G4INCLStandardPropagationModel.hh"
#include "G4INCLRandom.hh"
#include "G4INCLRanecu.hh"
#include "G4INCLFinalState.hh"
#include "G4INCLParticleTable.hh"
#include "G4INCLKinematicsUtils.hh"

// signal handler (for Linux and GCC)
#include "G4INCLSignalHandling.hh"

// For I/O
#include "IWriter.hh"
#include "ASCIIWriter.hh"
#include "ProtobufWriter.hh"
#include "INCLTree.hh"
#include "ROOTWriter.hh"
#include "HDF5Writer.hh"

// For configuration
#include "G4INCLConfig.hh"
#include "ConfigParser.hh"

// For logging
#include "G4INCLLogger.hh"

// Generic de-excitation interface
#include "G4INCLIDeExcitation.hh"

// ABLA v3p de-excitation
#ifdef INCL_DEEXCITATION_ABLAXX
#include "G4INCLAblaInterface.hh"
#endif

// ABLA07 de-excitation
#ifdef INCL_DEEXCITATION_ABLA07
#include "G4INCLAbla07Interface.hh"
#endif

// SMM de-excitation
#ifdef INCL_DEEXCITATION_SMM
#include "G4INCLSMMInterface.hh"
#endif

// GEMINIXX de-excitation
#ifdef INCL_DEEXCITATION_GEMINIXX
#include "G4INCLGEMINIXXInterface.hh"
#endif

#ifdef HAS_BOOST_DATE_TIME
#include <boost/date_time/posix_time/posix_time.hpp>
namespace bpt = boost::posix_time;
#endif

#ifdef HAS_BOOST_TIMER
#include <boost/timer/timer.hpp>
namespace bt = boost::timer;
#endif
using namespace G4INCL;
int main(int argc, char *argv[]) {
  ConfigParser theParser;
  G4INCL::Config *theConfig = theParser.parse(argc,argv);
  if(!theConfig)
    return 0;
#ifdef INCL_SIGNAL_HANDLING
  enableSignalHandling();
#endif

#ifdef INCL_HDF5_USE
  G4INCL::IO::INCLTree *theTreeHDF5 = new G4INCL::IO::INCLTree(new G4INCL::IO::HDF5Writer(), theConfig->getOutputFileRoot() + ".hdf5");
#endif
#ifdef INCL_ASCII_USE
  G4INCL::IO::INCLTree *theTreeASCII = new G4INCL::IO::INCLTree(new G4INCL::IO::ASCIIWriter(), theConfig->getOutputFileRoot() + ".out");
#endif
#ifdef INCL_PROTOBUF_USE
  G4INCL::IO::INCLTree *theTreeProtobuf = new G4INCL::IO::INCLTree(new G4INCL::IO::ProtobufWriter(), theConfig->getOutputFileRoot() + ".pb");
#endif
#ifdef INCL_ROOT_USE
  G4INCL::IO::INCLTree *theTreeROOT =
    new G4INCL::IO::INCLTree(new G4INCL::IO::ROOTWriter(theConfig->getROOTSelectionString(),
                                                        theConfig->getConciseROOTTree(),
                                                        theConfig->getInverseKinematics()),
                             theConfig->getOutputFileRoot() + ".root");
#endif

  // The G4INCL:: -prefix in front of class names means that all INCL
  // related classes are in G4INCL -namespace.
  G4INCL::INCL *theINCLModel = new G4INCL::INCL(theConfig);

  // Echo the input parameters to the log file
  INCL_INFO_NOCOMMENT(theParser.echo(theConfig) << std::endl);

  G4INCL::IDeExcitation *theDeExcitation = 0;
  switch(theConfig->getDeExcitationType()) {
#ifdef INCL_DEEXCITATION_ABLAXX
    case G4INCL::DeExcitationABLAv3p:
      theDeExcitation = new G4INCLAblaInterface(theConfig);
      break;
#endif
#ifdef INCL_DEEXCITATION_ABLA07
    case G4INCL::DeExcitationABLA07:
      theDeExcitation = new ABLA07CXX::Abla07Interface(theConfig);
      break;
#endif
#ifdef INCL_DEEXCITATION_SMM
    case G4INCL::DeExcitationSMM:
      theDeExcitation = new SMMCXX::SMMInterface(theConfig);
      break;
#endif
#ifdef INCL_DEEXCITATION_GEMINIXX
    case G4INCL::DeExcitationGEMINIXX:
      theDeExcitation = new G4INCLGEMINIXXInterface(theConfig);
      break;
#endif
    default:
      std::stringstream ss;
      ss << "########################################################\n"
         << "###              WARNING WARNING WARNING             ###\n"
         << "###                                                  ###\n"
         << "### You are running the code without any coupling to ###\n"
         << "###              a de-excitation model!              ###\n"
         << "###    Results will be INCOMPLETE and UNPHYSICAL!    ###\n"
         << "###    Are you sure this is what you want to do?     ###\n"
         << "########################################################\n";
      INCL_INFO(ss.str());
      std::cout << ss.str();
      break;
  }

  INCL_INFO(theConfig->summary() << std::endl);

#ifdef INCL_SIGNAL_HANDLING
  INCL_INFO("Signal handling activated" << std::endl);
#endif

#if defined(INCL_CACHING_CLUSTERING_MODEL_INTERCOMPARISON_HashMask)
  INCL_INFO("G4ClusteringModelIntercomparison will use HashMask caching" << std::endl);
#elif defined(INCL_CACHING_CLUSTERING_MODEL_INTERCOMPARISON_Set)
  INCL_INFO("G4ClusteringModelIntercomparison will use Set caching" << std::endl);
#elif defined(INCL_CACHING_CLUSTERING_MODEL_INTERCOMPARISON_None)
  INCL_INFO("G4ClusteringModelIntercomparison will not use any caching" << std::endl);
#endif

#if defined(INCL_AVATAR_SEARCH_FullSort)
  INCL_INFO("Avatar search will use FullSort algorithm" << std::endl);
#elif defined(INCL_AVATAR_SEARCH_MinElement)
  INCL_INFO("Avatar search will use MinElement algorithm" << std::endl);
#endif

  if(theConfig->getInverseKinematics()) {
    INCL_INFO("Will populate inverse-kinematics variables" << std::endl);
  }

  if(theConfig->getConciseROOTTree()) {
    INCL_INFO("Will produce a concise ROOT tree" << std::endl);
  }

  int eventToBeLogged = theConfig->getVerboseEvent(); // roughly corresponds to Alain's "ibert" variable
  int nShots = theConfig->getNumberOfShots();

  long lastAutoSaveEvent = -1;
  const long autoSaveFrequency = theConfig->getAutosaveFrequency();

  std::stringstream progressSS;
#ifdef HAS_BOOST_DATE_TIME
  bpt::ptime lapTime, endTime, etaTime;
  bpt::ptime startTime = bpt::microsec_clock::local_time();
  INCL_INFO("Event loop started at local time: " << startTime << std::endl);
#endif
#ifdef HAS_BOOST_TIMER
  std::stringstream occupationSS;
  bt::cpu_timer theTimer;
#endif
  G4INCL::Random::SeedVector const theInitialSeeds = G4INCL::Random::getSeeds();

  for(int i = 0; i < nShots; i++) {
    if(i == eventToBeLogged) G4INCL::Logger::setVerbosityLevel(10);

    G4INCL::Random::saveSeeds();

    INCL_DEBUG("Starting event " << i << std::endl);
    INCL_DEBUG("Random seeds " << G4INCL::Random::getSeeds() << std::endl);

    G4INCL::EventInfo result;
    result = theINCLModel->processEvent();
    result.event = i;

    INCL_DEBUG("End of event " << i << std::endl);
    if( result.transparent ) {
      INCL_DEBUG("Transparent event" << std::endl);
    } else {
      INCL_DEBUG("Number of produced particles: " << result.nParticles << std::endl);
      if(theDeExcitation != 0) {
	theDeExcitation->deExcite(&result);
	INCL_DEBUG("Number of produced particles after de-excitation: " << result.nParticles << std::endl);
      }

      // Fill the inverse kinematics variables, if appropriate
      if(theConfig->getInverseKinematics()) {
        double gamma = G4INCL::KinematicsUtils::gammaFromKineticEnergy(theConfig->getProjectileSpecies(), theConfig->getProjectileKineticEnergy());
        result.fillInverseKinematics(gamma);
      }

#ifdef INCL_HDF5_USE
        theTreeHDF5->fillEventInfo(result);
#endif
#ifdef INCL_ASCII_USE
        theTreeASCII->fillEventInfo(result);
#endif
#ifdef INCL_PROTOBUF_USE
	theTreeProtobuf->fillEventInfo(result);
#endif
#ifdef INCL_ROOT_USE
        theTreeROOT->fillEventInfo(result);
#endif
    }
    if(i == eventToBeLogged) G4INCL::Logger::setVerbosityLevel(theConfig->getVerbosity());

    // If we caught a SIGINT or a SIGTERM, leave the event loop and exit cleanly
#ifdef INCL_SIGNAL_HANDLING
    if(INTTERMsignalled) {
      INCL_INFO("Leaving event loop at the end of event " << i << std::endl);
      break;
    }
#endif

    // If we caught a SIGHUP or if we have gone too long without writing, flush
    // log and output files
    if(
#ifdef INCL_SIGNAL_HANDLING
       HUPsignalled ||
#endif
       i-lastAutoSaveEvent>=autoSaveFrequency) {
      const float progress = 100.*((float)(i+1))/((float)nShots);
      progressSS.str("");
      progressSS.precision(1);
      progressSS << std::fixed << progress;
      INCL_INFO("Flushing log and output files at the end of event " << i << ", " << progressSS.str() << "% done" << std::endl
             << "  random seeds: " << G4INCL::Random::getSeeds() << std::endl);
#ifdef HAS_BOOST_DATE_TIME
      lapTime = bpt::microsec_clock::local_time();
      const bpt::time_duration elapsedTime = lapTime - startTime;
      const long elapsedMillisecs = elapsedTime.total_milliseconds();
      const long remainingMillisecs = long(elapsedMillisecs*double(nShots-i-1)/double(i+1));
      const bpt::time_duration remainingDuration = bpt::milliseconds(remainingMillisecs);
      etaTime = lapTime + remainingDuration;
      INCL_INFO("  local time: " << lapTime << std::endl
             << "  remaining: " << remainingDuration << ", ETA: " << etaTime << std::endl);
#endif
#ifdef HAS_BOOST_TIMER
      const bt::cpu_times elapsedCPU = theTimer.elapsed();
      const double CPUOccupation = 100.*((double)(elapsedCPU.system+elapsedCPU.user))/((double)elapsedCPU.wall);
      occupationSS.str("");
      occupationSS.precision(1);
      occupationSS << std::fixed << CPUOccupation;
      INCL_INFO("  average CPU occupation: " << occupationSS.str() << "%" << std::endl);
#endif
      theINCLModel->finalizeGlobalInfo(theInitialSeeds);
#if defined(INCL_HDF5_USE) || defined(INCL_ASCII_USE) || defined(INCL_PROTOBUF_USE) || defined(INCL_ROOT_USE)
      G4INCL::GlobalInfo globalResult = theINCLModel->getGlobalInfo();
      lastAutoSaveEvent = i;
#endif
      G4INCL::Logger::flush();
#ifdef INCL_HDF5_USE
      theTreeHDF5->fillGlobalInfo(globalResult);
      theTreeHDF5->flush();
#endif
#ifdef INCL_ASCII_USE
      theTreeASCII->fillGlobalInfo(globalResult);
      theTreeASCII->flush();
#endif
#ifdef INCL_PROTOBUF_USE
      theTreeProtobuf->fillGlobalInfo(globalResult);
      theTreeProtobuf->flush();
#endif
#ifdef INCL_ROOT_USE
      theTreeROOT->fillGlobalInfo(globalResult);
      theTreeROOT->flush();
#endif
#ifdef INCL_SIGNAL_HANDLING
      HUPsignalled = false;
#endif
    }
  }

  // Fill in the global results
  theINCLModel->finalizeGlobalInfo(theInitialSeeds);
  G4INCL::GlobalInfo globalResult = theINCLModel->getGlobalInfo();
#ifdef INCL_HDF5_USE
  theTreeHDF5->fillGlobalInfo(globalResult);
  delete theTreeHDF5;
#endif
#ifdef INCL_ASCII_USE
  theTreeASCII->fillGlobalInfo(globalResult);
  delete theTreeASCII;
#endif
#ifdef INCL_PROTOBUF_USE
  theTreeProtobuf->fillGlobalInfo(globalResult);
  delete theTreeProtobuf;
#endif
#ifdef INCL_ROOT_USE
  theTreeROOT->fillGlobalInfo(globalResult);
  delete theTreeROOT;
#endif

  INCL_INFO("##################\n" << "### End of run ###\n"  << "##################\n" << std::endl);
#ifdef HAS_BOOST_DATE_TIME
  endTime = bpt::microsec_clock::local_time();
  INCL_INFO("Calculation ended at " << endTime << std::endl);
  INCL_INFO("Elapsed time = " << endTime - startTime << std::endl);
#endif
  INCL_INFO("Final random seeds = " << G4INCL::Random::getSeeds() << std::endl);
#ifdef INCL_COUNT_RND_CALLS
  unsigned long long nCalls = G4INCL::Random::getNumberOfCalls();
  const double nCallsPerReaction = (globalResult.nShots-globalResult.nTransparents!=0) ?
    ((double)nCalls)/((double)(globalResult.nShots-globalResult.nTransparents))
    : 0.;
  const double nCallsPerEvent = ((double)nCalls)/((double)(globalResult.nShots));
  INCL_INFO("Number of calls to RNG = " << nCalls
            << " (" << nCallsPerReaction << " per reaction, "
            << nCallsPerEvent << " per event)" << std::endl);
#endif
  INCL_INFO("Number of shots = " << globalResult.nShots << std::endl);
  INCL_INFO("Reaction cross section (mb) = "
      << globalResult.reactionCrossSection
      << " +- " << globalResult.errorReactionCrossSection << std::endl);
  INCL_INFO("Forced-compound-nucleus cross section (mb) = "
      << globalResult.forcedCNCrossSection
      << " +- " << globalResult.errorForcedCNCrossSection << std::endl);
  INCL_INFO("Complete-fusion cross section (mb) = "
      << globalResult.completeFusionCrossSection
      << " +- " << globalResult.errorCompleteFusionCrossSection << std::endl);
  const double ratio = (globalResult.nShots-globalResult.nTransparents!=0) ?
    ((double)globalResult.nEnergyViolationInteraction)/((double)(globalResult.nShots-globalResult.nTransparents))
    : 0.;
  INCL_INFO("Energy-violation cross section for binary collisions/decays (mb) = "
      << globalResult.energyViolationInteractionCrossSection
      << " (" << ratio
      << " per reaction)" << std::endl);
  INCL_INFO("Geometrical cross section (mb) = " << globalResult.geometricCrossSection << std::endl);
  if(theDeExcitation != 0) {
    delete theDeExcitation;
  }
  delete theINCLModel;

  return 0;
}
