/** \file G4INCLEventInfo.cc
 * \brief Simple container for output of event results.
 *
 * Contains the results of an INCL cascade.
 *
 * \date 21 January 2011
 * \author Davide Mancusi
 */

#include "G4INCLEventInfo.hh"
#include "G4INCLGlobals.hh"
#include "G4INCLParticleTable.hh"
#include <cmath>

namespace G4INCL {

  G4ThreadLocal Int_t EventInfo::eventNumber = 0;

  void EventInfo::fillInverseKinematics(const Double_t gamma) {
    const Double_t beta = std::sqrt(1.-1./(gamma*gamma));
    for(Int_t i=0; i<nParticles; ++i) {
      // determine the particle mass from the kinetic energy and the momentum;
      // this ensures consistency with the masses uses by the models
      Double_t mass;
      if(EKin[i]>0.) {
        mass = std::max(
                        0.5 * (px[i]*px[i]+py[i]*py[i]+pz[i]*pz[i]-EKin[i]*EKin[i]) / EKin[i],
                        0.0);
      } else {
        INCL_WARN("Particle with null kinetic energy in fillInverseKinematics, cannot determine its mass:\n"
                  << "  A=" << A[i] << ", Z=" << Z[i] << '\n'
                  << "  EKin=" << EKin[i] << ", px=" << px[i] << ", py=" << py[i] << ", pz=" << pz[i] << '\n'
                  << "  Falling back to the mass from the INCL ParticleTable" << '\n');
        mass = ParticleTable::getRealMass(A[i], Z[i]);
      }

      const Double_t ETot = EKin[i] + mass;
      const Double_t ETotPrime = gamma*(ETot - beta*pz[i]);
      EKinPrime[i] = ETotPrime - mass;
      pzPrime[i] = -gamma*(pz[i] - beta*ETot);
      const Double_t pPrime = std::sqrt(px[i]*px[i] + py[i]*py[i] + pzPrime[i]*pzPrime[i]);
      const Double_t cosThetaPrime = (pPrime>0.) ? (pzPrime[i]/pPrime) : 1.;
      if(cosThetaPrime>=1.)
        thetaPrime[i] = 0.;
      else if(cosThetaPrime<=-1.)
        thetaPrime[i] = 180.;
      else
        thetaPrime[i] = Math::toDegrees(Math::arcCos(cosThetaPrime));
    }
  }

  void EventInfo::remnantToParticle(const int remnantIndex) {
    A[nParticles] = ARem[remnantIndex];
    Z[nParticles] = ZRem[remnantIndex];
    emissionTime[nParticles] = stoppingTime;

    px[nParticles] = pxRem[remnantIndex];
    py[nParticles] = pyRem[remnantIndex];
    pz[nParticles] = pzRem[remnantIndex];

    const double plab = std::sqrt(pxRem[remnantIndex]*pxRem[remnantIndex]
                                  +pyRem[remnantIndex]*pyRem[remnantIndex]
                                  +pzRem[remnantIndex]*pzRem[remnantIndex]);
    double pznorm = pzRem[remnantIndex]/plab;
    if(pznorm>1.)
      pznorm = 1.;
    else if(pznorm<-1.)
      pznorm = -1.;
    theta[nParticles] = Math::toDegrees(Math::arcCos(pznorm));
    phi[nParticles] = Math::toDegrees(std::atan2(pyRem[remnantIndex],pxRem[remnantIndex]));

    EKin[nParticles] = EKinRem[remnantIndex];
    origin[nParticles] = -1; // Origin: cascade
    history.push_back(""); // history
    nParticles++;
    assert(history.size()==(unsigned int)nParticles);
  }
}

