/** \file G4INCLIFunction1D.cc
 * \brief Functor for 1-dimensional mathematical functions
 *
 * \date 16 July 2011
 * \author Davide Mancusi
 */

#include <algorithm>
#include <cmath>
#include <cstdlib>
#include "G4INCLIFunction1D.hh"
#include "G4INCLLogger.hh"
#include "G4INCLInvFInterpolationTable.hh"

namespace G4INCL {

  const double IFunction1D::integrationCoefficients[] = {
      2.*95.0/288.0,
      317.0/240.0,
      23.0/30.0,
      793.0/720.0,
      157.0/160.0,
      157.0/160.0,
      793.0/720.0,
      23.0/30.0,
      317.0/240.0,
  };

  double IFunction1D::integrate(const double x0, const double x1, const double step) const {
    double xi = std::max(x0, xMin);
    double xa = std::min(x1, xMax);
    double sign;

    if(x1 <= x0) {
      sign = -1.0;
      std::swap(xi, xa);
    } else
      sign = 1.0;

    const double interval = xa - xi;

    int nIntervals;
    if(step<0.) {
      nIntervals = 45;
    } else {
      nIntervals = int(interval/step);

      // Round up nIntervals to the closest multiple of 9
      int remainder = nIntervals % 9;
      if (remainder != 0)
        nIntervals += 9 - remainder;

      nIntervals = std::max(nIntervals, 9);
    }

    const double dx = interval/nIntervals;
    double result = (operator()(xi) + operator()(xa)) * integrationCoefficients[0]/2;
    for(int j = 1; j<nIntervals; ++j) {
      const double x = xi + interval*double(j)/double(nIntervals);
      const unsigned index = j%9;
      result += operator()(x) * integrationCoefficients[index];
    }

    return result*dx*sign;

  }

  IFunction1D *IFunction1D::primitive() const {
    class Primitive : public IFunction1D {
      public:
        Primitive(IFunction1D const * const f) :
          IFunction1D(f->getXMinimum(), f->getXMaximum()),
          theFunction(f)
      {}

        double operator()(const double x) const {
          return theFunction->integrate(xMin,x);
        }
      private:
        IFunction1D const * const theFunction;
    } *thePrimitive = new Primitive(this);

    return thePrimitive;
  }

  InterpolationTable *IFunction1D::inverseCDFTable(IFunction1D::ManipulatorFunc fWrap, const int nNodes) const {
    class InverseCDF : public IFunction1D {
      public:
        InverseCDF(IFunction1D const * const f, ManipulatorFunc fw) :
          IFunction1D(f->getXMinimum(), f->getXMaximum()),
          theFunction(f),
          normalisation(1./theFunction->integrate(xMin,xMax)),
          fWrap(fw)
      {}

        double operator()(const double x) const {
          if(fWrap)
            return fWrap(std::min(1., normalisation * theFunction->integrate(xMin,x)));
          else
            return std::min(1., normalisation * theFunction->integrate(xMin,x));
        }
      private:
        IFunction1D const * const theFunction;
        const double normalisation;
        ManipulatorFunc fWrap;
    } *theInverseCDF = new InverseCDF(this, fWrap);

    InterpolationTable *theTable = new InvFInterpolationTable(*theInverseCDF, nNodes);
    delete theInverseCDF;
    return theTable;
  }
}

