/** \file G4INCLInterpolationTable.cc
 * \brief Simple interpolation table
 *
 * \date 30 January 2014
 * \author Davide Mancusi
 */

#include <cassert>
#include <algorithm>
#include <functional>
#include "G4INCLInterpolationTable.hh"

namespace G4INCL {

  InterpolationTable::InterpolationTable() : IFunction1D() {}

  InterpolationTable::InterpolationTable(std::vector<double> const &x, std::vector<double> const &y)
    : IFunction1D(x.front(), x.back())
  {
    assert(x.size()==y.size());
    // Assert that the x vector is sorted
    assert(std::adjacent_find(nodes.begin(), nodes.end(), std::greater<InterpolationNode>()) == nodes.end());

    for(unsigned i = 0; i < x.size(); ++i)
      nodes.push_back(InterpolationNode(x.at(i), y.at(i), 0.));

    initDerivatives();
  }

  std::vector<double> InterpolationTable::getNodeAbscissae() const {
    std::vector<double> x(nodes.size());
    std::transform(nodes.begin(), nodes.end(), x.begin(),
                   std::mem_fun_ref(&InterpolationNode::getX));
    return x;
  }

  std::vector<double> InterpolationTable::getNodeValues() const {
    std::vector<double> y(nodes.size());
    std::transform(nodes.begin(), nodes.end(), y.begin(),
                   std::mem_fun_ref(&InterpolationNode::getY));
    return y;
  }

  void InterpolationTable::initDerivatives() {
    for(unsigned i = 0; i < nodes.size()-1; i++) {
      if((nodes.at(i+1).getX() - nodes.at(i).getX()) == 0.0) // Safeguard against division by zero
        nodes[i].setYPrime(0.0);
      else
        nodes[i].setYPrime((nodes.at(i+1).getY() - nodes.at(i).getY())/(nodes.at(i+1).getX() - nodes.at(i).getX()));
    }
    nodes.back().setYPrime(nodes.at(nodes.size()-2).getYPrime()); // Duplicate the last value
  }

  double InterpolationTable::operator()(const double x) const {
    // Find the relevant interpolation bin
    InterpolationNode xNode(x,0.,0.);
    std::vector<InterpolationNode>::const_iterator iter =
      std::lower_bound(nodes.begin(), nodes.end(), xNode);

    if(iter==nodes.begin())
      return nodes.front().getY();

    if(iter==nodes.end())
      return nodes.back().getY();

    std::vector<InterpolationNode>::const_iterator previousIter = iter - 1;
    const double dx = x - previousIter->getX();
    return previousIter->getY() + previousIter->getYPrime()*dx;
  }

  std::string InterpolationTable::print() const {
    std::string message;
    for(std::vector<InterpolationNode>::const_iterator n=nodes.begin(), e=nodes.end(); n!=e; ++n)
      message += n->print();
    return message;
  }

}
