/*
 * G4INCLRandom.cc
 *
 *  \date 7 June 2009
 * \author Pekka Kaitaniemi
 */

#include "G4INCLRandom.hh"
#include "G4INCLGlobals.hh"
#include <cassert>

#include "G4INCLRanecu.hh"
#include "G4INCLRanecu3.hh"
#include "G4INCLGeant4Random.hh"
#include "G4INCLLogger.hh"

namespace G4INCL {

  namespace Random {

    namespace {

      G4ThreadLocal IRandomGenerator* theGenerator = NULL;

#ifdef INCL_COUNT_RND_CALLS
      G4ThreadLocal unsigned long long nCalls;
#endif

      G4ThreadLocal SeedVector *savedSeeds = NULL;

      G4ThreadLocal Adapter *theAdapter = NULL;

    }

    void setGenerator(G4INCL::IRandomGenerator *aGenerator) {
      if(isInitialized()) {
        INCL_ERROR("INCL random number generator already initialized." << '\n');
      } else {
#ifdef INCL_COUNT_RND_CALLS
        nCalls = 0;
#endif
        theGenerator = aGenerator;
      }
      if(!theAdapter)
        theAdapter = new Adapter();
    }

    void setSeeds(const SeedVector &sv) {
      theGenerator->setSeeds(sv);
    }

    SeedVector getSeeds() {
      return theGenerator->getSeeds();
    }

    double shoot() {
#ifdef INCL_COUNT_RND_CALLS
      nCalls++;
#endif
      return theGenerator->flat();
    }

    double shoot0() {
      double r;
      while( (r=shoot()) <= 0. ) /* Loop checking, 10.07.2015, D.Mancusi */
        ;
      return r;
    }

    double shoot1() {
      double r;
      while( (r=shoot()) >= 1. ) /* Loop checking, 10.07.2015, D.Mancusi */
        ;
      return r;
    }

    double gauss(double sigma) {
      // generate a Gaussian random number with standard deviation sigma
      // uses the flat() and flat0() methods
      static G4ThreadLocal bool generated = false;
      static G4ThreadLocal double u, v;

      if( !generated )
      {
        u = shoot0();
        v = Math::twoPi*shoot();
        generated = true;
        return sigma*std::sqrt(-2*std::log(u))*std::cos(v);
      }
      else
      {
        generated = false;
        return sigma*std::sqrt(-2*std::log(u))*std::sin(v);
      }
    }

    ThreeVector normVector(double norm) {

      const double ctheta = (1.-2.*shoot());
      const double stheta = std::sqrt(1.-ctheta*ctheta);
      const double phi = Math::twoPi*shoot();
      return ThreeVector(
                         norm * stheta * std::cos(phi),
                         norm * stheta * std::sin(phi),
                         norm * ctheta);

    }

    ThreeVector sphereVector(double rmax) {
      return normVector( rmax*Math::pow13(shoot0()) );
    }

    ThreeVector gaussVector(double sigma) {
      const double sigmax = sigma * Math::oneOverSqrtThree;
      return ThreeVector(gauss(sigmax), gauss(sigmax), gauss(sigmax));
    }

    std::pair<double,double> correlatedGaussian(const double corrCoeff, const double x0, const double sigma) {
      assert(corrCoeff<=1. && corrCoeff>=-1.);
      double factor = 1.-corrCoeff*corrCoeff;
      if(factor<=0.)
        factor=0.;
      const double x = gauss(sigma) + x0;
      const double y = corrCoeff * x + gauss(sigma*std::sqrt(factor)) + x0;
      return std::make_pair(x, y);
    }

    std::pair<double,double> correlatedUniform(const double corrCoeff) {
      std::pair<double,double> gaussians = correlatedGaussian(corrCoeff);
      return std::make_pair(Math::gaussianCDF(gaussians.first), Math::gaussianCDF(gaussians.second));
    }

    void deleteGenerator() {
      delete theGenerator;
      theGenerator = NULL;
      delete savedSeeds;
      savedSeeds = NULL;
      delete theAdapter;
      theAdapter = NULL;
    }

    bool isInitialized() {
      if(theGenerator == 0) return false;
      return true;
    }

#ifdef INCL_COUNT_RND_CALLS
    /// \brief Return the number of calls to the RNG
    unsigned long long getNumberOfCalls() {
      return nCalls;
    }
#endif

    void saveSeeds() {
      if(!savedSeeds)
        savedSeeds = new SeedVector;

      (*savedSeeds) = theGenerator->getSeeds();
    }

    SeedVector getSavedSeeds() {
      if(!savedSeeds)
        savedSeeds = new SeedVector;

      return *savedSeeds;
    }

    void initialize(Config const * const
#ifndef INCLXX_IN_GEANT4_MODE
                    theConfig
#endif
                    ) {
#ifdef INCLXX_IN_GEANT4_MODE
      Random::setGenerator(new Geant4RandomGenerator());
#else // INCLXX_IN_GEANT4_MODE
      RNGType rng = theConfig->getRNGType();
      if(rng == RanecuType)
        setGenerator(new Ranecu(theConfig->getRandomSeeds()));
      else if(rng == Ranecu3Type)
        setGenerator(new Ranecu3(theConfig->getRandomSeeds()));
      else
        setGenerator(NULL);
#endif // INCLXX_IN_GEANT4_MODE
    }

    int Adapter::operator()(const int n) const {
      return shootInteger(n);
    }

    Adapter const &getAdapter() {
      return *theAdapter;
    }

  }

}
