/** \file G4INCLRootFinder.hh
 * \brief Static root-finder algorithm.
 *
 * Provides a stateless root-finder algorithm.
 *
 * \date 2nd March 2011
 * \author Davide Mancusi
 */

#include "G4INCLRootFinder.hh"
#include "G4INCLGlobals.hh"
#include "G4INCLLogger.hh"
#include <utility>
#include <cmath>

namespace G4INCL {

  namespace RootFinder {

    namespace {

      /// \brief Tolerance on the y value
      const double toleranceY = 1.e-4;

      /// \brief Maximum number of iterations for convergence
      const int maxIterations=50;

      /** \brief Bracket the root of the function f.
       *
       * Tries to find a bracketing value for the function root.
       *
       * \param f pointer to a RootFunctor
       * \param x0 starting value
       * \return if the root could be bracketed, returns two values of x
       *   bracketing the root, as a pair. If the bracketing failed, returns a
       *   pair with first > second.
       */
      std::pair<double,double> bracketRoot(RootFunctor const * const f, double x0) {
        double y0 = (*f)(x0);

        const double scaleFactor = 1.5;

        double x1;
        if(x0!=0.)
          x1=scaleFactor*x0;
        else
          x1=1.;
        double y1 = (*f)(x1);

        if(Math::sign(y0)!=Math::sign(y1))
          return std::make_pair(x0,x1);

        const double scaleFactorMinus1 = 1./scaleFactor;
        double oldx0, oldx1, oldy1;
        int iterations=0;
        do {
          if(iterations > maxIterations) {
            INCL_DEBUG("Could not bracket the root." << '\n');
            return std::make_pair((double) 1.,(double) -1.);
          }

          oldx0=x0;
          oldx1=x1;
          oldy1=y1;

          x0 *= scaleFactorMinus1;
          x1 *= scaleFactor;
          y0 = (*f)(x0);
          y1 = (*f)(x1);
          iterations++;
        } while(Math::sign(y0)==Math::sign(y1)); /* Loop checking, 10.07.2015, D.Mancusi */

        if(Math::sign(y1)==Math::sign(oldy1))
          return std::make_pair(x0,oldx0);
        else
          return std::make_pair(oldx1,x1);
      }

    }

    Solution solve(RootFunctor const * const f, const double x0) {
      // If we already have the solution, do nothing
      const double y0 = (*f)(x0);
      if( std::abs(y0) < toleranceY ) {
        return Solution(x0,y0);
      }

      // Bracket the root and set the initial values
      std::pair<double,double> bracket = bracketRoot(f,x0);
      double x1 = bracket.first;
      double x2 = bracket.second;
      // If x1>x2, it means that we could not bracket the root. Return false.
      if(x1>x2) {
        // Maybe zero is a good solution?
        double y_at_zero = (*f)(0.);
        if(std::abs(y_at_zero)<=toleranceY) {
          f->cleanUp(true);
          return Solution(0.,y_at_zero);
        } else {
          INCL_DEBUG("Root-finding algorithm could not bracket the root." << '\n');
          f->cleanUp(false);
          return Solution();
        }
      }

      double y1 = (*f)(x1);
      double y2 = (*f)(x2);
      double x = x1;
      double y = y1;

      /* ********************************
       * Start of the false-position loop
       * ********************************/

      // Keep track of the last updated interval end (-1=left, 1=right)
      int lastUpdated = 0;

      for(int iterations=0; std::abs(y) > toleranceY; iterations++) {

        if(iterations > maxIterations) {
          INCL_DEBUG("Root-finding algorithm did not converge." << '\n');
          f->cleanUp(false);
          return Solution();
        }

        // Estimate the root position by linear interpolation
        x = (y1*x2-y2*x1)/(y1-y2);

        // Update the value of the function
        y = (*f)(x);

        // Update the bracketing interval
        if(Math::sign(y) == Math::sign(y1)) {
          x1=x;
          y1=y;
          if(lastUpdated==-1) y2 *= 0.5;
          lastUpdated = -1;
        } else {
          x2=x;
          y2=y;
          if(lastUpdated==1) y1 *= 0.5;
          lastUpdated = 1;
        }
      }

      /* ******************************
       * End of the false-position loop
       * ******************************/

      f->cleanUp(true);
      return Solution(x,y);
    }

  } // namespace RootFinder
}
